from . import ec
from .base import EmbitKey, EmbitError
from .misc import copy, const, secp256k1
from .networks import NETWORKS
from . import base58
from . import hashes
import hmac
from binascii import hexlify

HARDENED_INDEX = const(0x80000000)


class HDError(EmbitError):
    pass


class HDKey(EmbitKey):
    """HD Private or Public key"""

    def __init__(
        self,
        key: EmbitKey,  # more specifically, PrivateKey or PublicKey
        chain_code: bytes,
        version=None,
        depth: int = 0,
        fingerprint: bytes = b"\x00\x00\x00\x00",
        child_number: int = 0,
    ):
        self.key = key
        if len(key.serialize()) != 32 and len(key.serialize()) != 33:
            raise HDError("Invalid key. Should be private or compressed public")
        if version is not None:
            self.version = version
        else:
            if len(key.serialize()) == 32:
                self.version = NETWORKS["main"]["xprv"]
            else:
                self.version = NETWORKS["main"]["xpub"]
        self.chain_code = chain_code
        self.depth = depth
        self.fingerprint = fingerprint
        self._my_fingerprint = b""
        self.child_number = child_number
        # check that base58[1:4] is "prv" or "pub"
        if self.is_private and self.to_base58()[1:4] != "prv":
            raise HDError("Invalid version")
        if not self.is_private and self.to_base58()[1:4] != "pub":
            raise HDError("Invalid version")

    @classmethod
    def from_seed(cls, seed: bytes, version=NETWORKS["main"]["xprv"]):
        """Creates a root private key from 64-byte seed"""
        raw = hmac.new(b"Bitcoin seed", seed, digestmod="sha512").digest()
        private_key = ec.PrivateKey(raw[:32])
        chain_code = raw[32:]
        return cls(private_key, chain_code, version=version)

    @classmethod
    def from_base58(cls, s: str):
        b = base58.decode_check(s)
        return cls.parse(b)

    @property
    def my_fingerprint(self) -> bytes:
        if not self._my_fingerprint:
            sec = self.sec()
            self._my_fingerprint = hashes.hash160(sec)[:4]
        return self._my_fingerprint

    @property
    def is_private(self) -> bool:
        """checks if the HDKey is private or public"""
        return self.key.is_private

    @property
    def secret(self):
        if not self.is_private:
            raise HDError("Key is not private")
        return self.key.secret

    def write_to(self, stream, version=None) -> int:
        if version is None:
            version = self.version
        res = stream.write(version)
        res += stream.write(bytes([self.depth]))
        res += stream.write(self.fingerprint)
        res += stream.write(self.child_number.to_bytes(4, "big"))
        res += stream.write(self.chain_code)
        if self.is_private:
            res += stream.write(b"\x00")
        res += stream.write(self.key.serialize())
        return res

    def to_base58(self, version=None) -> str:
        b = self.serialize(version)
        res = base58.encode_check(b)
        if res[1:4] == "prv" and not self.is_private:
            raise HDError("Invalid version for private key")
        if res[1:4] == "pub" and self.is_private:
            raise HDError("Invalid version for public key")
        return res

    @classmethod
    def from_string(cls, s: str):
        return cls.from_base58(s)

    def to_string(self, version=None):
        return self.to_base58(version)

    @classmethod
    def read_from(cls, stream):
        version = stream.read(4)
        depth = stream.read(1)[0]
        fingerprint = stream.read(4)
        child_number = int.from_bytes(stream.read(4), "big")
        chain_code = stream.read(32)
        k = stream.read(33)
        if k[0] == 0:
            key = ec.PrivateKey.parse(k[1:])
        else:
            key = ec.PublicKey.parse(k)

        if len(version) < 4 or len(fingerprint) < 4 or len(chain_code) < 32:
            raise HDError("Not enough bytes")
        hd = cls(
            key,
            chain_code,
            version=version,
            depth=depth,
            fingerprint=fingerprint,
            child_number=child_number,
        )
        subver = hd.to_base58()[1:4]
        if subver != "prv" and subver != "pub":
            raise HDError("Invalid version")
        if depth == 0 and child_number != 0:
            raise HDError("zero depth with non-zero index")
        if depth == 0 and fingerprint != b"\x00\x00\x00\x00":
            raise HDError("zero depth with non-zero parent")
        return hd

    def to_public(self, version=None):
        if not self.is_private:
            raise HDError("Already public")
        if version is None:
            # detect network
            for net in NETWORKS:
                for k in NETWORKS[net]:
                    if "prv" in k and NETWORKS[net][k] == self.version:
                        # xprv -> xpub, zprv -> zpub etc
                        version = NETWORKS[net][k.replace("prv", "pub")]
                        break
        if version is None:
            raise HDError("Can't find proper version. Provide it with version keyword")
        return self.__class__(
            self.key.get_public_key(),
            self.chain_code,
            version=version,
            depth=self.depth,
            fingerprint=self.fingerprint,
            child_number=self.child_number,
        )

    def get_public_key(self):
        return self.key.get_public_key() if self.is_private else self.key

    def sec(self) -> bytes:
        """Returns SEC serialization of the public key"""
        return self.key.sec()

    def xonly(self) -> bytes:
        return self.key.xonly()

    def taproot_tweak(self, h=b""):
        return HDKey(
            self.key.taproot_tweak(h),
            self.chain_code,
            version=self.version,
            depth=self.depth,
            fingerprint=self.fingerprint,
            child_number=self.child_number,
        )

    def child(self, index: int, hardened: bool = False):
        """Derives a child HDKey"""
        if index > 0xFFFFFFFF:
            raise HDError("Index should be less then 2^32")
        if hardened and index < HARDENED_INDEX:
            index += HARDENED_INDEX
        if index >= HARDENED_INDEX:
            hardened = True
        if hardened and not self.is_private:
            raise HDError("Can't do hardened with public key")

        # we need pubkey for fingerprint anyways
        sec = self.sec()
        fingerprint = hashes.hash160(sec)[:4]
        if hardened:
            data = b"\x00" + self.key.serialize() + index.to_bytes(4, "big")
        else:
            data = sec + index.to_bytes(4, "big")
        raw = hmac.new(self.chain_code, data, digestmod="sha512").digest()
        secret = raw[:32]
        chain_code = raw[32:]
        if self.is_private:
            secret = secp256k1.ec_privkey_add(secret, self.key.serialize())
            key = ec.PrivateKey(secret)
        else:
            # copy of internal secp256k1 point structure
            point = copy(self.key._point)
            point = secp256k1.ec_pubkey_add(point, secret)
            key = ec.PublicKey(point)
        return HDKey(
            key,
            chain_code,
            version=self.version,
            depth=self.depth + 1,
            fingerprint=fingerprint,
            child_number=index,
        )

    def derive(self, path):
        """path: int array or a string starting with m/"""
        if isinstance(path, str):
            # string of the form m/44h/0'/ind
            path = parse_path(path)
        child = self
        for idx in path:
            child = child.child(idx)
        return child

    def sign(self, msg_hash: bytes) -> ec.Signature:
        """signs a hash of the message with the private key"""
        if not self.is_private:
            raise HDError("HD public key can't sign")
        return self.key.sign(msg_hash)

    def schnorr_sign(self, msg_hash):
        if not self.is_private:
            raise HDError("HD public key can't sign")
        return self.key.schnorr_sign(msg_hash)

    def verify(self, sig, msg_hash) -> bool:
        return self.key.verify(sig, msg_hash)

    def schnorr_verify(self, sig, msg_hash) -> bool:
        return self.key.schnorr_verify(sig, msg_hash)

    def __eq__(self, other):
        # skip version
        return self.serialize()[4:] == other.serialize()[4:]

    def __hash__(self):
        return hash(self.serialize())


def detect_version(path, default="xprv", network=None) -> bytes:
    """
    Detects slip-132 version from the path for certain network.
    Trying to be smart, use if you want, but with care.
    """
    key = default
    net = network
    if network is None:
        net = NETWORKS["main"]
    if isinstance(path, str):
        path = parse_path(path)
    if len(path) == 0:
        return network[key]
    if path[0] == HARDENED_INDEX + 84:
        key = "z" + default[1:]
    elif path[0] == HARDENED_INDEX + 49:
        key = "y" + default[1:]
    elif path[0] == HARDENED_INDEX + 48:
        if len(path) >= 4:
            if path[3] == HARDENED_INDEX + 1:
                key = "Y" + default[1:]
            elif path[3] == HARDENED_INDEX + 2:
                key = "Z" + default[1:]
    if network is None and len(path) > 1 and path[1] == HARDENED_INDEX + 1:
        net = NETWORKS["test"]
    return net[key]


def _parse_der_item(e: str) -> int:
    if e[-1] in {"h", "H", "'"}:
        return int(e[:-1]) + HARDENED_INDEX
    else:
        return int(e)


def parse_path(path: str) -> list:
    """converts derivation path of the form m/44h/1'/0'/0/32 to int array"""
    arr = path.rstrip("/").split("/")
    if arr[0] == "m":
        arr = arr[1:]
    if len(arr) == 0:
        return []
    return [_parse_der_item(e) for e in arr]


def path_to_str(path: list, fingerprint=None) -> str:
    s = "m" if fingerprint is None else hexlify(fingerprint).decode()
    for el in path:
        if el >= HARDENED_INDEX:
            s += "/%dh" % (el - HARDENED_INDEX)
        else:
            s += "/%d" % el
    return s



from bitcoin_client.ledger_bitcoin.ripemd import ripemd160


def test_ripemd160():
    """RIPEMD-160 test vectors."""
    # See https://homes.esat.kuleuven.be/~bosselae/ripemd160.html
    for msg, hexout in [
        (b"", "9c1185a5c5e9fc54612808977ee8f548b2258d31"),
        (b"a", "0bdc9d2d256b3ee9daae347be6f4dc835a467ffe"),
        (b"abc", "8eb208f7e05d987a9b044a8e98c6b087f15a0bfc"),
        (b"message digest", "5d0689ef49d2fae572b881b123a85ffa21595f36"),
        (b"abcdefghijklmnopqrstuvwxyz", "f71c27109c692c1b56bbdceb5b9d2865b3708dbc"),
        (
            b"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
            "12a053384a9c0c88e405a06c27dcf49ada62eb2b",
        ),
        (
            b"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",
            "b0e20b6e3116640286ed3a87a5713079b21f5189",
        ),
        (b"1234567890" * 8, "9b752e45573d4b39f4dbd3323cab82bf63326bfb"),
        (b"a" * 1000000, "52783243c1697bdbe16d37f97f68f08325dc1528"),
    ]:
        assert ripemd160(msg).hex() == hexout



import re

from enum import IntEnum
from typing import List

from hashlib import sha256

from .common import serialize_str, AddressType, write_varint
from .merkle import MerkleTree, element_hash

class WalletType(IntEnum):
    WALLET_POLICY_V1 = 1
    WALLET_POLICY_V2 = 2


# should not be instantiated directly
class WalletPolicyBase:
    def __init__(self, name: str, version: WalletType) -> None:
        self.name = name
        self.version = version

        if (version != WalletType.WALLET_POLICY_V1 and version != WalletType.WALLET_POLICY_V2):
            raise ValueError("Invalid wallet policy version")

    def serialize(self) -> bytes:
        return b"".join([
            self.version.value.to_bytes(1, byteorder="big"),
            serialize_str(self.name)
        ])

    @property
    def id(self) -> bytes:
        return sha256(self.serialize()).digest()


class WalletPolicy(WalletPolicyBase):
    """
    Represents a wallet stored with a wallet policy.
    For version V2, the wallet is serialized as follows:
       - 1 byte   : wallet version
       - 1 byte   : length of the wallet name (max 64)
       - (var)    : wallet name (ASCII string)
       - (varint) : length of the descriptor template
       - 32-bytes : sha256 hash of the descriptor template
       - (varint) : number of keys (not larger than 252)
       - 32-bytes : root of the Merkle tree of all the keys information.

    The specific format of the keys is deferred to subclasses.
    """

    def __init__(self, name: str, descriptor_template: str, keys_info: List[str], version: WalletType = WalletType.WALLET_POLICY_V2):
        super().__init__(name, version)
        self.descriptor_template = descriptor_template
        self.keys_info = keys_info

    @property
    def n_keys(self) -> int:
        return len(self.keys_info)

    def serialize(self) -> bytes:
        keys_info_hashes = map(lambda k: element_hash(k.encode()), self.keys_info)

        descriptor_template_sha256 = sha256(self.descriptor_template.encode()).digest()

        return b"".join([
            super().serialize(),
            write_varint(len(self.descriptor_template.encode())),
            self.descriptor_template.encode() if self.version == WalletType.WALLET_POLICY_V1 else descriptor_template_sha256,
            write_varint(len(self.keys_info)),
            MerkleTree(keys_info_hashes).root
        ])

    def get_descriptor(self, change: bool) -> str:
        desc = self.descriptor_template
        for i in reversed(range(self.n_keys)):
            key = self.keys_info[i]
            desc = desc.replace(f"@{i}", key)

        # in V1, /** is part of the key; in V2, it's part of the policy map. This handles either
        desc = desc.replace("/**", f"/{1 if change else 0}/*")

        if self.version == WalletType.WALLET_POLICY_V2:
            # V2, the /<M;N> syntax is supported. Replace with M if not change, or with N if change
            regex = r"/<(\d+);(\d+)>"
            desc = re.sub(regex, "/\\2" if change else "/\\1", desc)

        return desc

class MultisigWallet(WalletPolicy):
    def __init__(self, name: str, address_type: AddressType, threshold: int, keys_info: List[str], sorted: bool = True, version: WalletType = WalletType.WALLET_POLICY_V2) -> None:
        n_keys = len(keys_info)

        if not (1 <= threshold <= n_keys <= 16):
            raise ValueError("Invalid threshold or number of keys")

        multisig_op = "sortedmulti" if sorted else "multi"

        if (address_type == AddressType.LEGACY):
            policy_prefix = f"sh({multisig_op}("
            policy_suffix = f"))"
        elif address_type == AddressType.WIT:
            policy_prefix = f"wsh({multisig_op}("
            policy_suffix = f"))"
        elif address_type == AddressType.SH_WIT:
            policy_prefix = f"sh(wsh({multisig_op}("
            policy_suffix = f")))"
        else:
            raise ValueError(f"Unexpected address type: {address_type}")

        key_placeholder_suffix = "/**" if version == WalletType.WALLET_POLICY_V2 else ""

        descriptor_template = "".join([
            policy_prefix,
            str(threshold) + ",",
            ",".join("@" + str(l) + key_placeholder_suffix for l in range(n_keys)),
            policy_suffix
        ])

        super().__init__(name, descriptor_template, keys_info, version)

        self.threshold = threshold



from io import BytesIO
from typing import List, Optional, Literal
from enum import Enum
from typing import Union

import hashlib

UINT64_MAX: int = 18446744073709551615
UINT32_MAX: int = 4294967295
UINT16_MAX: int = 65535


# from bitcoin-core/HWI
class Chain(Enum):
    """
    The blockchain network to use
    """
    MAIN = 0 #: Bitcoin Main network
    TEST = 1 #: Bitcoin Test network
    REGTEST = 2 #: Bitcoin Core Regression Test network
    SIGNET = 3 #: Bitcoin Signet

    def __str__(self) -> str:
        return self.name.lower()

    def __repr__(self) -> str:
        return str(self)

    @staticmethod
    def argparse(s: str) -> Union['Chain', str]:
        try:
            return Chain[s.upper()]
        except KeyError:
            return s

# from bitcoin-core/HWI
class AddressType(Enum):
    """
    The type of address to use
    """
    LEGACY = 1 #: Legacy address type. P2PKH for single sig, P2SH for scripts.
    WIT = 2    #: Native segwit v0 address type. P2WPKH for single sig, P2WPSH for scripts.
    SH_WIT = 3 #: Nested segwit v0 address type. P2SH-P2WPKH for single sig, P2SH-P2WPSH for scripts.
    TAP = 4    #: Segwit v1 Taproot address type. P2TR always.

    def __str__(self) -> str:
        return self.name.lower()

    def __repr__(self) -> str:
        return str(self)

    @staticmethod
    def argparse(s: str) -> Union['AddressType', str]:
        try:
            return AddressType[s.upper()]
        except KeyError:
            return s

def bip32_path_from_string(path: str) -> List[bytes]:
    splitted_path: List[str] = path.split("/")

    if not splitted_path:
        raise Exception(f"BIP32 path format error: '{path}'")

    if "m" in splitted_path and splitted_path[0] == "m":
        splitted_path = splitted_path[1:]

    return [int(p).to_bytes(4, byteorder="big") if "'" not in p
            else (0x80000000 | int(p[:-1])).to_bytes(4, byteorder="big")
            for p in splitted_path]


def write_varint(n: int) -> bytes:
    if n <= 0xFC:
        return n.to_bytes(1, byteorder="little")

    if n <= UINT16_MAX:
        return b"\xFD" + n.to_bytes(2, byteorder="little")

    if n <= UINT32_MAX:
        return b"\xFE" + n.to_bytes(4, byteorder="little")

    if n <= UINT64_MAX:
        return b"\xFF" + n.to_bytes(8, byteorder="little")

    raise ValueError(f"Can't write to varint: '{n}'!")


def read_varint(buf: BytesIO,
                prefix: Optional[bytes] = None) -> int:
    b: bytes = prefix if prefix else buf.read(1)

    if not b:
        raise ValueError(f"Can't read prefix: '{b}'!")

    n: int = {b"\xfd": 2, b"\xfe": 4, b"\xff": 8}.get(b, 1)  # default to 1

    b = buf.read(n) if n > 1 else b

    if len(b) != n:
        raise ValueError("Can't read varint!")

    return int.from_bytes(b, byteorder="little")


def read_uint(buf: BytesIO,
              bit_len: int,
              byteorder: Literal['big', 'little'] = 'little') -> int:
    size: int = bit_len // 8
    b: bytes = buf.read(size)

    if len(b) < size:
        raise ValueError(f"Can't read u{bit_len} in buffer!")

    return int.from_bytes(b, byteorder)


def serialize_str(value: str) -> bytes:
    return len(value.encode()).to_bytes(1, byteorder="big") + value.encode()


def ripemd160(x: bytes) -> bytes:
    try:
        h = hashlib.new("ripemd160")
        h.update(x)
        return h.digest()
    except BaseException:
        # ripemd160 is not always present in hashlib.
        # Fallback to custom implementation if missing.
        from . import ripemd
        return ripemd.ripemd160(x)

def sha256(s: bytes) -> bytes:
    return hashlib.new('sha256', s).digest()


def hash160(s: bytes) -> bytes:
    return ripemd160(sha256(s))


def hash256(s: bytes) -> bytes:
    return sha256(sha256(s))


class ByteStreamParser:
    def __init__(self, input: bytes):
        self.stream = BytesIO(input)

    def assert_empty(self) -> bytes:
        if self.stream.read(1) != b'':
            raise ValueError("Byte stream was expected to be empty")

    def read_bytes(self, n: int) -> bytes:
        result = self.stream.read(n)
        if len(result) < n:
            raise ValueError("Byte stream exhausted")
        return result

    def read_uint(self, n: int, byteorder: Literal['big', 'little'] = "big") -> int:
        return int.from_bytes(self.read_bytes(n), byteorder)

    def read_varint(self) -> int:
        prefix = self.read_uint(1)

        if prefix == 253:
            return self.read_uint(2, 'little')
        elif prefix == 254:
            return self.read_uint(4, 'little')
        elif prefix == 255:
            return self.read_uint(8, 'little')
        else:
            return prefix



from io import BytesIO
from .. import script
from ..networks import NETWORKS
from .errors import DescriptorError
from .base import DescriptorBase
from .miniscript import Miniscript, Multi, Sortedmulti
from .arguments import Key
from .taptree import TapTree


class Descriptor(DescriptorBase):
    def __init__(
        self,
        miniscript=None,
        sh=False,
        wsh=True,
        key=None,
        wpkh=True,
        taproot=False,
        taptree=None,
    ):
        # TODO: add support for taproot scripts
        # Should:
        # - accept taptree without a key
        # - accept key without taptree
        # - raise if miniscript is not None, but taproot=True
        # - raise if taptree is not None, but taproot=False
        if key is None and miniscript is None and taptree is None:
            raise DescriptorError("Provide a key, miniscript or taptree")
        if miniscript is not None:
            # will raise if can't verify
            miniscript.verify()
            if miniscript.type != "B":
                raise DescriptorError("Top level miniscript should be 'B'")
            # check all branches have the same length
            branches = {
                len(k.branches) for k in miniscript.keys if k.branches is not None
            }
            if len(branches) > 1:
                raise DescriptorError("All branches should have the same length")
        self.sh = sh
        self.wsh = wsh
        self.key = key
        self.miniscript = miniscript
        self.wpkh = wpkh
        self.taproot = taproot
        self.taptree = taptree or TapTree()
        # make sure all keys are either taproot or not
        for k in self.keys:
            k.taproot = taproot

    @property
    def script_len(self):
        if self.taproot:
            return 34  # OP_1 <32:xonly>
        if self.miniscript:
            return len(self.miniscript)
        if self.wpkh:
            return 22  # 00 <20:pkh>
        return 25  # OP_DUP OP_HASH160 <20:pkh> OP_EQUALVERIFY OP_CHECKSIG

    @property
    def num_branches(self):
        return max([k.num_branches for k in self.keys])

    def branch(self, branch_index=None):
        if self.miniscript:
            return type(self)(
                self.miniscript.branch(branch_index),
                self.sh,
                self.wsh,
                None,
                self.wpkh,
                self.taproot,
            )
        else:
            return type(self)(
                None,
                self.sh,
                self.wsh,
                self.key.branch(branch_index),
                self.wpkh,
                self.taproot,
                self.taptree.branch(branch_index),
            )

    @property
    def is_wildcard(self):
        return any([key.is_wildcard for key in self.keys])

    @property
    def is_wrapped(self):
        return self.sh and self.is_segwit

    @property
    def is_legacy(self):
        return not (self.is_segwit or self.is_taproot)

    @property
    def is_segwit(self):
        return (
            (self.wsh and self.miniscript) or (self.wpkh and self.key) or self.taproot
        )

    @property
    def is_pkh(self):
        return self.key is not None and not self.taproot

    @property
    def is_taproot(self):
        return self.taproot

    @property
    def is_basic_multisig(self) -> bool:
        # TODO: should be true for taproot basic multisig with NUMS as internal key
        # Sortedmulti is subclass of Multi
        return bool(self.miniscript and isinstance(self.miniscript, Multi))

    @property
    def is_sorted(self) -> bool:
        return bool(self.is_basic_multisig and isinstance(self.miniscript, Sortedmulti))

    def scriptpubkey_type(self):
        if self.is_taproot:
            return "p2tr"
        if self.sh:
            return "p2sh"
        if self.is_pkh:
            if self.is_legacy:
                return "p2pkh"
            if self.is_segwit:
                return "p2wpkh"
        else:
            return "p2wsh"

    @property
    def brief_policy(self):
        if self.taptree:
            return "taptree"
        if self.key:
            return "single key"
        if self.is_basic_multisig:
            return (
                str(self.miniscript.args[0])
                + " of "
                + str(len(self.keys))
                + " multisig"
                + (" (sorted)" if self.is_sorted else "")
            )
        return "miniscript"

    @property
    def full_policy(self):
        if (self.key and not self.taptree) or self.is_basic_multisig:
            return self.brief_policy
        s = str(self.miniscript or self)
        for i, k in enumerate(self.keys):
            s = s.replace(str(k), chr(65 + i))
        return s

    def derive(self, idx, branch_index=None):
        if self.miniscript:
            return type(self)(
                self.miniscript.derive(idx, branch_index),
                self.sh,
                self.wsh,
                None,
                self.wpkh,
                self.taproot,
            )
        else:
            return type(self)(
                None,
                self.sh,
                self.wsh,
                self.key.derive(idx, branch_index),
                self.wpkh,
                self.taproot,
                self.taptree.derive(idx, branch_index),
            )

    def to_public(self):
        if self.miniscript:
            return type(self)(
                self.miniscript.to_public(),
                self.sh,
                self.wsh,
                None,
                self.wpkh,
                self.taproot,
            )
        else:
            return type(self)(
                None,
                self.sh,
                self.wsh,
                self.key.to_public(),
                self.wpkh,
                self.taproot,
                self.taptree.to_public(),
            )

    def owns(self, psbt_scope):
        """Checks if psbt input or output belongs to this descriptor"""
        # we can't check if we don't know script_pubkey
        if psbt_scope.script_pubkey is None:
            return False
        # quick check of script_pubkey type
        if psbt_scope.script_pubkey.script_type() != self.scriptpubkey_type():
            return False
        for pub, der in psbt_scope.bip32_derivations.items():
            # check of the fingerprints
            for k in self.keys:
                if not k.is_extended:
                    continue
                res = k.check_derivation(der)
                if res:
                    idx, branch_idx = res
                    sc = self.derive(idx, branch_index=branch_idx).script_pubkey()
                    # if derivation is found but scriptpubkey doesn't match - fail
                    return sc == psbt_scope.script_pubkey
        for pub, (leafs, der) in psbt_scope.taproot_bip32_derivations.items():
            # check of the fingerprints
            for k in self.keys:
                if not k.is_extended:
                    continue
                res = k.check_derivation(der)
                if res:
                    idx, branch_idx = res
                    sc = self.derive(idx, branch_index=branch_idx).script_pubkey()
                    # if derivation is found but scriptpubkey doesn't match - fail
                    return sc == psbt_scope.script_pubkey
        return False

    def check_derivation(self, derivation_path):
        for k in self.keys:
            # returns a tuple branch_idx, idx
            der = k.check_derivation(derivation_path)
            if der is not None:
                return der
        return None

    def witness_script(self):
        if self.wsh and self.miniscript is not None:
            return script.Script(self.miniscript.compile())

    def redeem_script(self):
        if not self.sh:
            return None
        if self.miniscript:
            if not self.wsh:
                return script.Script(self.miniscript.compile())
            else:
                return script.p2wsh(script.Script(self.miniscript.compile()))
        else:
            return script.p2wpkh(self.key)

    def script_pubkey(self):
        # covers sh-wpkh, sh and sh-wsh
        if self.taproot:
            return script.p2tr(self.key, self.taptree)
        if self.sh:
            return script.p2sh(self.redeem_script())
        if self.wsh:
            return script.p2wsh(self.witness_script())
        if self.miniscript:
            return script.Script(self.miniscript.compile())
        if self.wpkh:
            return script.p2wpkh(self.key)
        return script.p2pkh(self.key)

    def address(self, network=NETWORKS["main"]):
        return self.script_pubkey().address(network)

    @property
    def keys(self):
        if self.taptree and self.key:
            return [self.key] + self.taptree.keys
        elif self.taptree:
            return self.taptree.keys
        elif self.key:
            return [self.key]
        return self.miniscript.keys

    @classmethod
    def from_string(cls, desc):
        s = BytesIO(desc.encode())
        res = cls.read_from(s)
        left = s.read()
        if len(left) > 0 and not left.startswith(b"#"):
            raise DescriptorError("Unexpected characters after descriptor: %r" % left)
        return res

    @classmethod
    def read_from(cls, s):
        # starts with sh(wsh()), sh() or wsh()
        start = s.read(7)
        sh = False
        wsh = False
        wpkh = False
        is_miniscript = True
        taproot = False
        taptree = TapTree()
        if start.startswith(b"tr("):
            taproot = True
            s.seek(-4, 1)
        elif start.startswith(b"sh(wsh("):
            sh = True
            wsh = True
        elif start.startswith(b"wsh("):
            sh = False
            wsh = True
            s.seek(-3, 1)
        elif start.startswith(b"sh(wpkh"):
            is_miniscript = False
            sh = True
            wpkh = True
            assert s.read(1) == b"("
        elif start.startswith(b"wpkh("):
            is_miniscript = False
            wpkh = True
            s.seek(-2, 1)
        elif start.startswith(b"pkh("):
            is_miniscript = False
            s.seek(-3, 1)
        elif start.startswith(b"sh("):
            sh = True
            wsh = False
            s.seek(-4, 1)
        else:
            raise ValueError("Invalid descriptor (starts with '%s')" % start.decode())
        # taproot always has a key, and may have taptree miniscript
        if taproot:
            miniscript = None
            key = Key.read_from(s, taproot=True)
            nbrackets = 1
            c = s.read(1)
            # TODO: should it be ok to pass just taptree without a key?
            # check if we have taptree after the key
            if c != b",":
                s.seek(-1, 1)
            else:
                taptree = TapTree.read_from(s)
        elif is_miniscript:
            miniscript = Miniscript.read_from(s)
            key = None
            nbrackets = int(sh) + int(wsh)
        # single key for sure
        else:
            miniscript = None
            key = Key.read_from(s, taproot=taproot)
            nbrackets = 1 + int(sh)
        end = s.read(nbrackets)
        if end != b")" * nbrackets:
            raise ValueError(
                "Invalid descriptor (expected ')' but ends with '%s')" % end.decode()
            )
        return cls(
            miniscript,
            sh=sh,
            wsh=wsh,
            key=key,
            wpkh=wpkh,
            taproot=taproot,
            taptree=taptree,
        )

    def to_string(self):
        if self.taproot:
            if self.taptree:
                return "tr(%s,%s)" % (self.key, self.taptree)
            return "tr(%s)" % self.key
        if self.miniscript is not None:
            res = str(self.miniscript)
            if self.wsh:
                res = "wsh(%s)" % res
        else:
            if self.wpkh:
                res = "wpkh(%s)" % self.key
            else:
                res = "pkh(%s)" % self.key
        if self.sh:
            res = "sh(%s)" % res
        return res



"""
Original version: https://raw.githubusercontent.com/bitcoin-core/HWI/3fe369d0379212fae1c72729a179d133b0adc872/hwilib/_script.py
Distributed under the MIT License.

Bitcoin Script utilities
************************
"""

from typing import (
    Optional,
    Sequence,
    Tuple,
)


def is_opreturn(script: bytes) -> bool:
    """
    Determine whether a script is an OP_RETURN output script.

    :param script: The script
    :returns: Whether the script is an OP_RETURN output script
    """
    return script[0] == 0x6a


def is_p2sh(script: bytes) -> bool:
    """
    Determine whether a script is a P2SH output script.

    :param script: The script
    :returns: Whether the script is a P2SH output script
    """
    return len(script) == 23 and script[0] == 0xa9 and script[1] == 0x14 and script[22] == 0x87


def is_p2pkh(script: bytes) -> bool:
    """
    Determine whether a script is a P2PKH output script.

    :param script: The script
    :returns: Whether the script is a P2PKH output script
    """
    return len(script) == 25 and script[0] == 0x76 and script[1] == 0xa9 and script[2] == 0x14 and script[23] == 0x88 and script[24] == 0xac


def is_p2pk(script: bytes) -> bool:
    """
    Determine whether a script is a P2PK output script.

    :param script: The script
    :returns: Whether the script is a P2PK output script
    """
    return (len(script) == 35 or len(script) == 67) and (script[0] == 0x21 or script[0] == 0x41) and script[-1] == 0xac


def is_p2tr(script: bytes) -> bool:
    """
    Determine whether a script is a P2TR output script.

    :param script: The script
    :returns: Whether the script is a P2TR output script
    """
    return len(script) == 34 and script[0] == 0x51 and script[1] == 0x20


def is_witness(script: bytes) -> Tuple[bool, int, bytes]:
    """
    Determine whether a script is a segwit output script.
    If so, also returns the witness version and witness program.

    :param script: The script
    :returns: A tuple of a bool indicating whether the script is a segwit output script,
        an int representing the witness version,
        and the bytes of the witness program.
    """
    if len(script) < 4 or len(script) > 42:
        return (False, 0, b"")

    if script[0] != 0 and (script[0] < 81 or script[0] > 96):
        return (False, 0, b"")

    if script[1] + 2 == len(script):
        return (True, script[0] - 0x50 if script[0] else 0, script[2:])

    return (False, 0, b"")


def is_p2wpkh(script: bytes) -> bool:
    """
    Determine whether a script is a P2WPKH output script.

    :param script: The script
    :returns: Whether the script is a P2WPKH output script
    """
    is_wit, wit_ver, wit_prog = is_witness(script)
    if not is_wit:
        return False
    elif wit_ver != 0:
        return False
    return len(wit_prog) == 20


def is_p2wsh(script: bytes) -> bool:
    """
    Determine whether a script is a P2WSH output script.

    :param script: The script
    :returns: Whether the script is a P2WSH output script
    """
    is_wit, wit_ver, wit_prog = is_witness(script)
    if not is_wit:
        return False
    elif wit_ver != 0:
        return False
    return len(wit_prog) == 32


# Only handles up to 15 of 15. Returns None if this script is not a
# multisig script. Returns (m, pubkeys) otherwise.
def parse_multisig(script: bytes) -> Optional[Tuple[int, Sequence[bytes]]]:
    """
    Determine whether a script is a multisig script. If so, determine the parameters of that multisig.

    :param script: The script
    :returns: ``None`` if the script is not multisig.
        If multisig, returns a tuple of the number of signers required,
        and a sequence of public key bytes.
    """
    # Get m
    m = script[0] - 80
    if m < 1 or m > 15:
        return None

    # Get pubkeys
    pubkeys = []
    offset = 1
    while True:
        pubkey_len = script[offset]
        if pubkey_len != 33:
            break
        offset += 1
        pubkeys.append(script[offset:offset + 33])
        offset += 33

    # Check things at the end
    n = script[offset] - 80
    if n != len(pubkeys):
        return None
    offset += 1
    op_cms = script[offset]
    if op_cms != 174:
        return None

    return (m, pubkeys)



import ctypes, os
import ctypes.util
import platform
import threading

from ctypes import (
    cast,
    byref,
    c_char,
    c_byte,
    c_int,
    c_uint,
    c_char_p,
    c_size_t,
    c_void_p,
    c_uint64,
    create_string_buffer,
    CFUNCTYPE,
    POINTER,
)

_lock = threading.Lock()


# @locked decorator
def locked(func):
    def wrapper(*args, **kwargs):
        with _lock:
            return func(*args, **kwargs)

    return wrapper


# Flags to pass to context_create.
CONTEXT_VERIFY = 0b0100000001
CONTEXT_SIGN = 0b1000000001
CONTEXT_NONE = 0b0000000001

# Flags to pass to ec_pubkey_serialize
EC_COMPRESSED = 0b0100000010
EC_UNCOMPRESSED = 0b0000000010


def _copy(a: bytes) -> bytes:
    """Ugly copy that works everywhere incl micropython"""
    if len(a) == 0:
        return b""
    return a[:1] + a[1:]


def _find_library():
    library_path = None
    extension = ""
    if platform.system() == "Darwin":
        extension = ".dylib"
    elif platform.system() == "Linux":
        extension = ".so"
    elif platform.system() == "Windows":
        extension = ".dll"

    path = os.path.join(
        os.path.dirname(__file__),
        "prebuilt/libsecp256k1_%s_%s%s"
        % (platform.system().lower(), platform.machine().lower(), extension),
    )
    if os.path.isfile(path):
        return path
    # try searching
    if not library_path:
        library_path = ctypes.util.find_library("libsecp256k1")
    if not library_path:
        library_path = ctypes.util.find_library("secp256k1")
    # library search failed
    if not library_path:
        if platform.system() == "Linux" and os.path.isfile(
            "/usr/local/lib/libsecp256k1.so.0"
        ):
            library_path = "/usr/local/lib/libsecp256k1.so.0"
    return library_path


@locked
def _init(flags=(CONTEXT_SIGN | CONTEXT_VERIFY)):
    library_path = _find_library()
    # meh, can't find library
    if not library_path:
        raise RuntimeError(
            "Can't find libsecp256k1 library. Make sure to compile and install it."
        )

    secp256k1 = ctypes.cdll.LoadLibrary(library_path)

    secp256k1.secp256k1_context_create.argtypes = [c_uint]
    secp256k1.secp256k1_context_create.restype = c_void_p

    secp256k1.secp256k1_context_randomize.argtypes = [c_void_p, c_char_p]
    secp256k1.secp256k1_context_randomize.restype = c_int

    secp256k1.secp256k1_ec_seckey_verify.argtypes = [c_void_p, c_char_p]
    secp256k1.secp256k1_ec_seckey_verify.restype = c_int

    secp256k1.secp256k1_ec_privkey_negate.argtypes = [c_void_p, c_char_p]
    secp256k1.secp256k1_ec_privkey_negate.restype = c_int

    secp256k1.secp256k1_ec_pubkey_negate.argtypes = [c_void_p, c_char_p]
    secp256k1.secp256k1_ec_pubkey_negate.restype = c_int

    secp256k1.secp256k1_ec_privkey_tweak_add.argtypes = [c_void_p, c_char_p, c_char_p]
    secp256k1.secp256k1_ec_privkey_tweak_add.restype = c_int

    secp256k1.secp256k1_ec_privkey_tweak_mul.argtypes = [c_void_p, c_char_p, c_char_p]
    secp256k1.secp256k1_ec_privkey_tweak_mul.restype = c_int

    secp256k1.secp256k1_ec_pubkey_create.argtypes = [c_void_p, c_void_p, c_char_p]
    secp256k1.secp256k1_ec_pubkey_create.restype = c_int

    secp256k1.secp256k1_ec_pubkey_parse.argtypes = [c_void_p, c_char_p, c_char_p, c_int]
    secp256k1.secp256k1_ec_pubkey_parse.restype = c_int

    secp256k1.secp256k1_ec_pubkey_serialize.argtypes = [
        c_void_p,
        c_char_p,
        c_void_p,
        c_char_p,
        c_uint,
    ]
    secp256k1.secp256k1_ec_pubkey_serialize.restype = c_int

    secp256k1.secp256k1_ec_pubkey_tweak_add.argtypes = [c_void_p, c_char_p, c_char_p]
    secp256k1.secp256k1_ec_pubkey_tweak_add.restype = c_int

    secp256k1.secp256k1_ec_pubkey_tweak_mul.argtypes = [c_void_p, c_char_p, c_char_p]
    secp256k1.secp256k1_ec_pubkey_tweak_mul.restype = c_int

    secp256k1.secp256k1_ecdsa_signature_parse_compact.argtypes = [
        c_void_p,
        c_char_p,
        c_char_p,
    ]
    secp256k1.secp256k1_ecdsa_signature_parse_compact.restype = c_int

    secp256k1.secp256k1_ecdsa_signature_serialize_compact.argtypes = [
        c_void_p,
        c_char_p,
        c_char_p,
    ]
    secp256k1.secp256k1_ecdsa_signature_serialize_compact.restype = c_int

    secp256k1.secp256k1_ecdsa_signature_parse_der.argtypes = [
        c_void_p,
        c_char_p,
        c_char_p,
        c_uint,
    ]
    secp256k1.secp256k1_ecdsa_signature_parse_der.restype = c_int

    secp256k1.secp256k1_ecdsa_signature_serialize_der.argtypes = [
        c_void_p,
        c_char_p,
        c_void_p,
        c_char_p,
    ]
    secp256k1.secp256k1_ecdsa_signature_serialize_der.restype = c_int

    secp256k1.secp256k1_ecdsa_sign.argtypes = [
        c_void_p,
        c_char_p,
        c_char_p,
        c_char_p,
        c_void_p,
        c_char_p,
    ]
    secp256k1.secp256k1_ecdsa_sign.restype = c_int

    secp256k1.secp256k1_ecdsa_verify.argtypes = [c_void_p, c_char_p, c_char_p, c_char_p]
    secp256k1.secp256k1_ecdsa_verify.restype = c_int

    secp256k1.secp256k1_ec_pubkey_combine.argtypes = [
        c_void_p,
        c_char_p,
        c_void_p,
        c_size_t,
    ]
    secp256k1.secp256k1_ec_pubkey_combine.restype = c_int

    # ecdh
    try:
        secp256k1.secp256k1_ecdh.argtypes = [
            c_void_p,  # ctx
            c_char_p,  # output
            c_char_p,  # point
            c_char_p,  # scalar
            CFUNCTYPE,  # hashfp
            c_void_p,  # data
        ]
        secp256k1.secp256k1_ecdh.restype = c_int
    except:
        pass

    # schnorr sig
    try:
        secp256k1.secp256k1_xonly_pubkey_from_pubkey.argtypes = [
            c_void_p,  # ctx
            c_char_p,  # xonly pubkey
            POINTER(c_int),  # parity
            c_char_p,  # pubkey
        ]
        secp256k1.secp256k1_xonly_pubkey_from_pubkey.restype = c_int

        secp256k1.secp256k1_schnorrsig_verify.argtypes = [
            c_void_p,  # ctx
            c_char_p,  # sig
            c_char_p,  # msg
            c_char_p,  # pubkey
        ]
        secp256k1.secp256k1_schnorrsig_verify.restype = c_int

        secp256k1.secp256k1_schnorrsig_sign.argtypes = [
            c_void_p,  # ctx
            c_char_p,  # sig
            c_char_p,  # msg
            c_char_p,  # keypair
            c_void_p,  # nonce_function
            c_char_p,  # extra data
        ]
        secp256k1.secp256k1_schnorrsig_sign.restype = c_int

        secp256k1.secp256k1_keypair_create.argtypes = [
            c_void_p,  # ctx
            c_char_p,  # keypair
            c_char_p,  # secret
        ]
        secp256k1.secp256k1_keypair_create.restype = c_int
    except:
        pass

    # recoverable module
    try:
        secp256k1.secp256k1_ecdsa_sign_recoverable.argtypes = [
            c_void_p,
            c_char_p,
            c_char_p,
            c_char_p,
            c_void_p,
            c_void_p,
        ]
        secp256k1.secp256k1_ecdsa_sign_recoverable.restype = c_int

        secp256k1.secp256k1_ecdsa_recoverable_signature_parse_compact.argtypes = [
            c_void_p,
            c_char_p,
            c_char_p,
            c_int,
        ]
        secp256k1.secp256k1_ecdsa_recoverable_signature_parse_compact.restype = c_int

        secp256k1.secp256k1_ecdsa_recoverable_signature_serialize_compact.argtypes = [
            c_void_p,
            c_char_p,
            c_char_p,
            c_char_p,
        ]
        secp256k1.secp256k1_ecdsa_recoverable_signature_serialize_compact.restype = (
            c_int
        )

        secp256k1.secp256k1_ecdsa_recoverable_signature_convert.argtypes = [
            c_void_p,
            c_char_p,
            c_char_p,
        ]
        secp256k1.secp256k1_ecdsa_recoverable_signature_convert.restype = c_int

        secp256k1.secp256k1_ecdsa_recover.argtypes = [
            c_void_p,
            c_char_p,
            c_char_p,
            c_char_p,
        ]
        secp256k1.secp256k1_ecdsa_recover.restype = c_int
    except:
        pass

    # zkp modules
    try:
        # generator module
        secp256k1.secp256k1_generator_parse.argtypes = [c_void_p, c_char_p, c_char_p]
        secp256k1.secp256k1_generator_parse.restype = c_int

        secp256k1.secp256k1_generator_serialize.argtypes = [
            c_void_p,
            c_char_p,
            c_char_p,
        ]
        secp256k1.secp256k1_generator_serialize.restype = c_int

        secp256k1.secp256k1_generator_generate.argtypes = [c_void_p, c_char_p, c_char_p]
        secp256k1.secp256k1_generator_generate.restype = c_int

        secp256k1.secp256k1_generator_generate_blinded.argtypes = [
            c_void_p,
            c_char_p,
            c_char_p,
            c_char_p,
        ]
        secp256k1.secp256k1_generator_generate_blinded.restype = c_int

        # pederson commitments
        secp256k1.secp256k1_pedersen_commitment_parse.argtypes = [
            c_void_p,
            c_char_p,
            c_char_p,
        ]
        secp256k1.secp256k1_pedersen_commitment_parse.restype = c_int

        secp256k1.secp256k1_pedersen_commitment_serialize.argtypes = [
            c_void_p,
            c_char_p,
            c_char_p,
        ]
        secp256k1.secp256k1_pedersen_commitment_serialize.restype = c_int

        secp256k1.secp256k1_pedersen_commit.argtypes = [
            c_void_p,
            c_char_p,
            c_char_p,
            c_uint64,
            c_char_p,
        ]
        secp256k1.secp256k1_pedersen_commit.restype = c_int

        secp256k1.secp256k1_pedersen_blind_generator_blind_sum.argtypes = [
            c_void_p,  # const secp256k1_context* ctx,
            POINTER(c_uint64),  # const uint64_t *value,
            c_void_p,  # const unsigned char* const* generator_blind,
            c_void_p,  # unsigned char* const* blinding_factor,
            c_size_t,  # size_t n_total,
            c_size_t,  # size_t n_inputs
        ]
        secp256k1.secp256k1_pedersen_blind_generator_blind_sum.restype = c_int

        secp256k1.secp256k1_pedersen_verify_tally.argtypes = [
            c_void_p,
            c_void_p,
            c_size_t,
            c_void_p,
            c_size_t,
        ]
        secp256k1.secp256k1_pedersen_verify_tally.restype = c_int

        # rangeproof
        secp256k1.secp256k1_rangeproof_rewind.argtypes = [
            c_void_p,  # ctx
            c_char_p,  # vbf out
            POINTER(c_uint64),  # value out
            c_char_p,  # message out
            POINTER(c_uint64),  # msg out len
            c_char_p,  # nonce
            POINTER(c_uint64),  # min value
            POINTER(c_uint64),  # max value
            c_char_p,  # pedersen commitment
            c_char_p,  # range proof
            c_uint64,  # proof len
            c_char_p,  # extra commitment (scriptpubkey)
            c_uint64,  # extra len
            c_char_p,  # generator
        ]
        secp256k1.secp256k1_rangeproof_rewind.restype = c_int

        secp256k1.secp256k1_rangeproof_verify.argtypes = [
            c_void_p,  # ctx
            POINTER(c_uint64),  # min value
            POINTER(c_uint64),  # max value
            c_char_p,  # pedersen commitment
            c_char_p,  # proof
            c_uint64,  # proof len
            c_char_p,  # extra
            c_uint64,  # extra len
            c_char_p,  # generator
        ]
        secp256k1.secp256k1_rangeproof_verify.restype = c_int

        secp256k1.secp256k1_rangeproof_sign.argtypes = [
            c_void_p,  # ctx
            c_char_p,  # proof
            POINTER(c_uint64),  # plen
            c_uint64,  # min_value
            c_char_p,  # commit
            c_char_p,  # blind
            c_char_p,  # nonce
            c_int,  # exp
            c_int,  # min_bits
            c_uint64,  # value
            c_char_p,  # message
            c_uint64,  # msg_len
            c_char_p,  # extra_commit
            c_uint64,  # extra_commit_len
            c_char_p,  # gen
        ]
        secp256k1.secp256k1_rangeproof_sign.restype = c_int

        # musig
        secp256k1.secp256k1_musig_pubkey_combine.argtypes = [
            c_void_p,
            c_void_p,
            c_char_p,
            c_void_p,
            c_void_p,
            c_size_t,
        ]
        secp256k1.secp256k1_musig_pubkey_combine.restype = c_int

        # surjection proofs
        secp256k1.secp256k1_surjectionproof_initialize.argtypes = [
            c_void_p,  # const secp256k1_context* ctx,
            c_char_p,  # secp256k1_surjectionproof* proof,
            POINTER(c_size_t),  # size_t *input_index,
            c_void_p,  # c_char_p, # const secp256k1_fixed_asset_tag* fixed_input_tags,
            c_size_t,  # const size_t n_input_tags,
            c_size_t,  # const size_t n_input_tags_to_use,
            c_char_p,  # const secp256k1_fixed_asset_tag* fixed_output_tag,
            c_size_t,  # const size_t n_max_iterations,
            c_char_p,  # const unsigned char *random_seed32
        ]
        secp256k1.secp256k1_surjectionproof_initialize.restype = c_int

        secp256k1.secp256k1_surjectionproof_generate.argtypes = [
            c_void_p,  # const secp256k1_context* ctx,
            c_char_p,  # secp256k1_surjectionproof* proof,
            c_char_p,  # const secp256k1_generator* ephemeral_input_tags,
            c_size_t,  # size_t n_ephemeral_input_tags,
            c_char_p,  # const secp256k1_generator* ephemeral_output_tag,
            c_size_t,  # size_t input_index,
            c_char_p,  # const unsigned char *input_blinding_key,
            c_char_p,  # const unsigned char *output_blinding_key
        ]
        secp256k1.secp256k1_surjectionproof_generate.restype = c_int

        secp256k1.secp256k1_surjectionproof_verify.argtypes = [
            c_void_p,  # const secp256k1_context* ctx,
            c_char_p,  # const secp256k1_surjectionproof* proof,
            c_char_p,  # const secp256k1_generator* ephemeral_input_tags,
            c_size_t,  # size_t n_ephemeral_input_tags,
            c_char_p,  # const secp256k1_generator* ephemeral_output_tag
        ]
        secp256k1.secp256k1_surjectionproof_verify.restype = c_int

        secp256k1.secp256k1_surjectionproof_serialize.argtypes = [
            c_void_p,  # const secp256k1_context* ctx,
            c_char_p,  # unsigned char *output,
            POINTER(c_size_t),  # size_t *outputlen,
            c_char_p,  # const secp256k1_surjectionproof *proof
        ]
        secp256k1.secp256k1_surjectionproof_serialize.restype = c_int

        secp256k1.secp256k1_surjectionproof_serialized_size.argtypes = [
            c_void_p,  # const secp256k1_context* ctx,
            c_char_p,  # const secp256k1_surjectionproof* proof
        ]
        secp256k1.secp256k1_surjectionproof_serialized_size.restype = c_size_t

        secp256k1.secp256k1_surjectionproof_parse.argtypes = [
            c_void_p,
            c_char_p,
            c_char_p,
            c_size_t,
        ]
        secp256k1.secp256k1_surjectionproof_parse.restype = c_int

    except:
        pass

    secp256k1.ctx = secp256k1.secp256k1_context_create(flags)

    r = secp256k1.secp256k1_context_randomize(secp256k1.ctx, os.urandom(32))

    return secp256k1


_secp = _init()


# bindings equal to ones in micropython
@locked
def context_randomize(seed, context=_secp.ctx):
    if len(seed) != 32:
        raise ValueError("Seed should be 32 bytes long")
    if _secp.secp256k1_context_randomize(context, seed) == 0:
        raise RuntimeError("Failed to randomize context")


@locked
def ec_pubkey_create(secret, context=_secp.ctx):
    if len(secret) != 32:
        raise ValueError("Private key should be 32 bytes long")
    pub = bytes(64)
    r = _secp.secp256k1_ec_pubkey_create(context, pub, secret)
    if r == 0:
        raise ValueError("Invalid private key")
    return pub


@locked
def ec_pubkey_parse(sec, context=_secp.ctx):
    if len(sec) != 33 and len(sec) != 65:
        raise ValueError("Serialized pubkey should be 33 or 65 bytes long")
    if len(sec) == 33:
        if sec[0] != 0x02 and sec[0] != 0x03:
            raise ValueError("Compressed pubkey should start with 0x02 or 0x03")
    else:
        if sec[0] != 0x04:
            raise ValueError("Uncompressed pubkey should start with 0x04")
    pub = bytes(64)
    r = _secp.secp256k1_ec_pubkey_parse(context, pub, sec, len(sec))
    if r == 0:
        raise ValueError("Failed parsing public key")
    return pub


@locked
def ec_pubkey_serialize(pubkey, flag=EC_COMPRESSED, context=_secp.ctx):
    if len(pubkey) != 64:
        raise ValueError("Pubkey should be 64 bytes long")
    if flag not in [EC_COMPRESSED, EC_UNCOMPRESSED]:
        raise ValueError("Invalid flag")
    sec = bytes(33) if (flag == EC_COMPRESSED) else bytes(65)
    sz = c_size_t(len(sec))
    r = _secp.secp256k1_ec_pubkey_serialize(context, sec, byref(sz), pubkey, flag)
    if r == 0:
        raise ValueError("Failed to serialize pubkey")
    return sec


@locked
def ecdsa_signature_parse_compact(compact_sig, context=_secp.ctx):
    if len(compact_sig) != 64:
        raise ValueError("Compact signature should be 64 bytes long")
    sig = bytes(64)
    r = _secp.secp256k1_ecdsa_signature_parse_compact(context, sig, compact_sig)
    if r == 0:
        raise ValueError("Failed parsing compact signature")
    return sig


@locked
def ecdsa_signature_parse_der(der, context=_secp.ctx):
    sig = bytes(64)
    r = _secp.secp256k1_ecdsa_signature_parse_der(context, sig, der, len(der))
    if r == 0:
        raise ValueError("Failed parsing compact signature")
    return sig


@locked
def ecdsa_signature_serialize_der(sig, context=_secp.ctx):
    if len(sig) != 64:
        raise ValueError("Signature should be 64 bytes long")
    der = bytes(78)  # max
    sz = c_size_t(len(der))
    r = _secp.secp256k1_ecdsa_signature_serialize_der(context, der, byref(sz), sig)
    if r == 0:
        raise ValueError("Failed serializing der signature")
    return der[: sz.value]


@locked
def ecdsa_signature_serialize_compact(sig, context=_secp.ctx):
    if len(sig) != 64:
        raise ValueError("Signature should be 64 bytes long")
    ser = bytes(64)
    r = _secp.secp256k1_ecdsa_signature_serialize_compact(context, ser, sig)
    if r == 0:
        raise ValueError("Failed serializing der signature")
    return ser


@locked
def ecdsa_signature_normalize(sig, context=_secp.ctx):
    if len(sig) != 64:
        raise ValueError("Signature should be 64 bytes long")
    sig2 = bytes(64)
    r = _secp.secp256k1_ecdsa_signature_normalize(context, sig2, sig)
    return sig2


@locked
def ecdsa_verify(sig, msg, pub, context=_secp.ctx):
    if len(sig) != 64:
        raise ValueError("Signature should be 64 bytes long")
    if len(msg) != 32:
        raise ValueError("Message should be 32 bytes long")
    if len(pub) != 64:
        raise ValueError("Public key should be 64 bytes long")
    r = _secp.secp256k1_ecdsa_verify(context, sig, msg, pub)
    return bool(r)


@locked
def ecdsa_sign(msg, secret, nonce_function=None, extra_data=None, context=_secp.ctx):
    if len(msg) != 32:
        raise ValueError("Message should be 32 bytes long")
    if len(secret) != 32:
        raise ValueError("Secret key should be 32 bytes long")
    if extra_data and len(extra_data) != 32:
        raise ValueError("Extra data should be 32 bytes long")
    sig = bytes(64)
    r = _secp.secp256k1_ecdsa_sign(
        context, sig, msg, secret, nonce_function, extra_data
    )
    if r == 0:
        raise ValueError("Failed to sign")
    return sig


@locked
def ec_seckey_verify(secret, context=_secp.ctx):
    if len(secret) != 32:
        raise ValueError("Secret should be 32 bytes long")
    return bool(_secp.secp256k1_ec_seckey_verify(context, secret))


@locked
def ec_privkey_negate(secret, context=_secp.ctx):
    if len(secret) != 32:
        raise ValueError("Secret should be 32 bytes long")
    b = _copy(secret)
    _secp.secp256k1_ec_privkey_negate(context, b)
    return b


@locked
def ec_pubkey_negate(pubkey, context=_secp.ctx):
    if len(pubkey) != 64:
        raise ValueError("Pubkey should be a 64-byte structure")
    pub = _copy(pubkey)
    r = _secp.secp256k1_ec_pubkey_negate(context, pub)
    if r == 0:
        raise ValueError("Failed to negate pubkey")
    return pub


@locked
def ec_privkey_tweak_add(secret, tweak, context=_secp.ctx):
    if len(secret) != 32 or len(tweak) != 32:
        raise ValueError("Secret and tweak should both be 32 bytes long")
    t = _copy(tweak)
    if _secp.secp256k1_ec_privkey_tweak_add(context, secret, tweak) == 0:
        raise ValueError("Failed to tweak the secret")
    return None


@locked
def ec_pubkey_tweak_add(pub, tweak, context=_secp.ctx):
    if len(pub) != 64:
        raise ValueError("Public key should be 64 bytes long")
    if len(tweak) != 32:
        raise ValueError("Tweak should be 32 bytes long")
    t = _copy(tweak)
    if _secp.secp256k1_ec_pubkey_tweak_add(context, pub, tweak) == 0:
        raise ValueError("Failed to tweak the public key")
    return None


@locked
def ec_privkey_add(secret, tweak, context=_secp.ctx):
    if len(secret) != 32 or len(tweak) != 32:
        raise ValueError("Secret and tweak should both be 32 bytes long")
    # ugly copy that works in mpy and py
    s = _copy(secret)
    t = _copy(tweak)
    if _secp.secp256k1_ec_privkey_tweak_add(context, s, t) == 0:
        raise ValueError("Failed to tweak the secret")
    return s


@locked
def ec_pubkey_add(pub, tweak, context=_secp.ctx):
    if len(pub) != 64:
        raise ValueError("Public key should be 64 bytes long")
    if len(tweak) != 32:
        raise ValueError("Tweak should be 32 bytes long")
    p = _copy(pub)
    if _secp.secp256k1_ec_pubkey_tweak_add(context, p, tweak) == 0:
        raise ValueError("Failed to tweak the public key")
    return p


@locked
def ec_privkey_tweak_mul(secret, tweak, context=_secp.ctx):
    if len(secret) != 32 or len(tweak) != 32:
        raise ValueError("Secret and tweak should both be 32 bytes long")
    if _secp.secp256k1_ec_privkey_tweak_mul(context, secret, tweak) == 0:
        raise ValueError("Failed to tweak the secret")


@locked
def ec_pubkey_tweak_mul(pub, tweak, context=_secp.ctx):
    if len(pub) != 64:
        raise ValueError("Public key should be 64 bytes long")
    if len(tweak) != 32:
        raise ValueError("Tweak should be 32 bytes long")
    if _secp.secp256k1_ec_pubkey_tweak_mul(context, pub, tweak) == 0:
        raise ValueError("Failed to tweak the public key")


@locked
def ec_pubkey_combine(*args, context=_secp.ctx):
    pub = bytes(64)
    pubkeys = (c_char_p * len(args))(*args)
    r = _secp.secp256k1_ec_pubkey_combine(context, pub, pubkeys, len(args))
    if r == 0:
        raise ValueError("Failed to combine pubkeys")
    return pub


# ecdh
@locked
def ecdh(pubkey, scalar, hashfn=None, data=None, context=_secp.ctx):
    if not len(pubkey) == 64:
        raise ValueError("Pubkey should be 64 bytes long")
    if not len(scalar) == 32:
        raise ValueError("Scalar should be 32 bytes long")
    secret = bytes(32)
    if hashfn is None:
        res = _secp.secp256k1_ecdh(context, secret, pubkey, scalar, None, None)
    else:

        def _hashfn(out, x, y):
            x = ctypes.string_at(x, 32)
            y = ctypes.string_at(y, 32)
            try:
                res = hashfn(x, y, data)
            except Exception as e:
                return 0
            out = cast(out, POINTER(c_char * 32))
            out.contents.value = res
            return 1

        HASHFN = CFUNCTYPE(c_int, c_void_p, c_void_p, c_void_p)
        res = _secp.secp256k1_ecdh(
            context, secret, pubkey, scalar, HASHFN(_hashfn), data
        )
    if res != 1:
        raise RuntimeError("Failed to compute the shared secret")
    return secret


# schnorrsig
@locked
def xonly_pubkey_from_pubkey(pubkey, context=_secp.ctx):
    if len(pubkey) != 64:
        raise ValueError("Pubkey should be 64 bytes long")
    pointer = POINTER(c_int)
    parity = pointer(c_int(0))
    xonly_pub = bytes(64)
    res = _secp.secp256k1_xonly_pubkey_from_pubkey(context, xonly_pub, parity, pubkey)
    if res != 1:
        raise RuntimeError("Failed to convert the pubkey")
    return xonly_pub, bool(parity.contents.value)


@locked
def schnorrsig_verify(sig, msg, pubkey, context=_secp.ctx):
    assert len(sig) == 64
    assert len(msg) == 32
    assert len(pubkey) == 64
    res = _secp.secp256k1_schnorrsig_verify(context, sig, msg, pubkey)
    return bool(res)


@locked
def keypair_create(secret, context=_secp.ctx):
    assert len(secret) == 32
    keypair = bytes(96)
    r = _secp.secp256k1_keypair_create(context, keypair, secret)
    if r == 0:
        raise ValueError("Failed to create keypair")
    return keypair


# not @locked because it uses keypair_create inside
def schnorrsig_sign(
    msg, keypair, nonce_function=None, extra_data=None, context=_secp.ctx
):
    assert len(msg) == 32
    if len(keypair) == 32:
        keypair = keypair_create(keypair, context=context)
    with _lock:
        assert len(keypair) == 96
        sig = bytes(64)
        r = _secp.secp256k1_schnorrsig_sign(
            context, sig, msg, keypair, nonce_function, extra_data
        )
        if r == 0:
            raise ValueError("Failed to sign")
        return sig


# recoverable
@locked
def ecdsa_sign_recoverable(msg, secret, context=_secp.ctx):
    if len(msg) != 32:
        raise ValueError("Message should be 32 bytes long")
    if len(secret) != 32:
        raise ValueError("Secret key should be 32 bytes long")
    sig = bytes(65)
    r = _secp.secp256k1_ecdsa_sign_recoverable(context, sig, msg, secret, None, None)
    if r == 0:
        raise ValueError("Failed to sign")
    return sig


@locked
def ecdsa_recoverable_signature_serialize_compact(sig, context=_secp.ctx):
    if len(sig) != 65:
        raise ValueError("Recoverable signature should be 65 bytes long")
    ser = bytes(64)
    idx = bytes(1)
    r = _secp.secp256k1_ecdsa_recoverable_signature_serialize_compact(
        context, ser, idx, sig
    )
    if r == 0:
        raise ValueError("Failed serializing der signature")
    return ser, idx[0]


@locked
def ecdsa_recoverable_signature_parse_compact(compact_sig, recid, context=_secp.ctx):
    if len(compact_sig) != 64:
        raise ValueError("Signature should be 64 bytes long")
    sig = bytes(65)
    r = _secp.secp256k1_ecdsa_recoverable_signature_parse_compact(
        context, sig, compact_sig, recid
    )
    if r == 0:
        raise ValueError("Failed parsing compact signature")
    return sig


@locked
def ecdsa_recoverable_signature_convert(sigin, context=_secp.ctx):
    if len(sigin) != 65:
        raise ValueError("Recoverable signature should be 65 bytes long")
    sig = bytes(64)
    r = _secp.secp256k1_ecdsa_recoverable_signature_convert(context, sig, sigin)
    if r == 0:
        raise ValueError("Failed converting signature")
    return sig


@locked
def ecdsa_recover(sig, msghash, context=_secp.ctx):
    if len(sig) != 65:
        raise ValueError("Recoverable signature should be 65 bytes long")
    if len(msghash) != 32:
        raise ValueError("Message should be 32 bytes long")
    pub = bytes(64)
    r = _secp.secp256k1_ecdsa_recover(context, pub, sig, msghash)
    if r == 0:
        raise ValueError("Failed to recover public key")
    return pub


# zkp modules


@locked
def pedersen_commitment_parse(inp, context=_secp.ctx):
    if len(inp) != 33:
        raise ValueError("Serialized commitment should be 33 bytes long")
    commit = bytes(64)
    r = _secp.secp256k1_pedersen_commitment_parse(context, commit, inp)
    if r == 0:
        raise ValueError("Failed to parse commitment")
    return commit


@locked
def pedersen_commitment_serialize(commit, context=_secp.ctx):
    if len(commit) != 64:
        raise ValueError("Commitment should be 64 bytes long")
    sec = bytes(33)
    r = _secp.secp256k1_pedersen_commitment_serialize(context, sec, commit)
    if r == 0:
        raise ValueError("Failed to serialize commitment")
    return sec


@locked
def pedersen_commit(vbf, value, gen, context=_secp.ctx):
    if len(gen) != 64:
        raise ValueError("Generator should be 64 bytes long")
    if len(vbf) != 32:
        raise ValueError(f"Blinding factor should be 32 bytes long, not {len(vbf)}")
    commit = bytes(64)
    r = _secp.secp256k1_pedersen_commit(context, commit, vbf, value, gen)
    if r == 0:
        raise ValueError("Failed to create commitment")
    return commit


@locked
def pedersen_blind_generator_blind_sum(
    values, gens, vbfs, num_inputs, context=_secp.ctx
):
    vals = (c_uint64 * len(values))(*values)
    vbf = bytes(vbfs[-1])
    p = c_char_p(vbf)  # obtain a pointer of various types
    address = cast(p, c_void_p).value

    vbfs_joined = (c_char_p * len(vbfs))(*vbfs[:-1], address)
    gens_joined = (c_char_p * len(gens))(*gens)
    res = _secp.secp256k1_pedersen_blind_generator_blind_sum(
        context, vals, gens_joined, vbfs_joined, len(values), num_inputs
    )
    if res == 0:
        raise ValueError("Failed to get the last blinding factor.")
    res = (c_char * 32).from_address(address).raw
    assert len(res) == 32
    return res


@locked
def pedersen_verify_tally(ins, outs, context=_secp.ctx):
    in_ptr = (c_char_p * len(ins))(*ins)
    out_ptr = (c_char_p * len(outs))(*outs)
    res = _secp.secp256k1_pedersen_verify_tally(
        context, in_ptr, len(in_ptr), out_ptr, len(out_ptr)
    )
    return bool(res)


# generator
@locked
def generator_parse(inp, context=_secp.ctx):
    if len(inp) != 33:
        raise ValueError("Serialized generator should be 33 bytes long")
    gen = bytes(64)
    r = _secp.secp256k1_generator_parse(context, gen, inp)
    if r == 0:
        raise ValueError("Failed to parse generator")
    return gen


@locked
def generator_generate(asset, context=_secp.ctx):
    if len(asset) != 32:
        raise ValueError("Asset should be 32 bytes long")
    gen = bytes(64)
    r = _secp.secp256k1_generator_generate(context, gen, asset)
    if r == 0:
        raise ValueError("Failed to generate generator")
    return gen


@locked
def generator_generate_blinded(asset, abf, context=_secp.ctx):
    if len(asset) != 32:
        raise ValueError("Asset should be 32 bytes long")
    if len(abf) != 32:
        raise ValueError("Asset blinding factor should be 32 bytes long")
    gen = bytes(64)
    r = _secp.secp256k1_generator_generate_blinded(context, gen, asset, abf)
    if r == 0:
        raise ValueError("Failed to generate generator")
    return gen


@locked
def generator_serialize(generator, context=_secp.ctx):
    if len(generator) != 64:
        raise ValueError("Generator should be 64 bytes long")
    sec = bytes(33)
    if _secp.secp256k1_generator_serialize(context, sec, generator) == 0:
        raise RuntimeError("Failed to serialize generator")
    return sec


# rangeproof
@locked
def rangeproof_rewind(
    proof,
    nonce,
    value_commitment,
    script_pubkey,
    generator,
    message_length=64,
    context=_secp.ctx,
):
    if len(generator) != 64:
        raise ValueError("Generator should be 64 bytes long")
    if len(nonce) != 32:
        raise ValueError("Nonce should be 32 bytes long")
    if len(value_commitment) != 64:
        raise ValueError("Value commitment should be 64 bytes long")

    pointer = POINTER(c_uint64)

    msg = b"\x00" * message_length
    msglen = pointer(c_uint64(len(msg)))

    vbf_out = b"\x00" * 32
    value_out = pointer(c_uint64(0))
    min_value = pointer(c_uint64(0))
    max_value = pointer(c_uint64(0))
    res = _secp.secp256k1_rangeproof_rewind(
        context,
        vbf_out,
        value_out,
        msg,
        msglen,
        nonce,
        min_value,
        max_value,
        value_commitment,
        proof,
        len(proof),
        script_pubkey,
        len(script_pubkey),
        generator,
    )
    if res != 1:
        raise RuntimeError("Failed to rewind the proof")
    return (
        value_out.contents.value,
        vbf_out,
        msg[: msglen.contents.value],
        min_value.contents.value,
        max_value.contents.value,
    )


# rangeproof


@locked
def rangeproof_verify(
    proof, value_commitment, script_pubkey, generator, context=_secp.ctx
):
    if len(generator) != 64:
        raise ValueError("Generator should be 64 bytes long")
    if len(value_commitment) != 64:
        raise ValueError("Value commitment should be 64 bytes long")

    pointer = POINTER(c_uint64)
    min_value = pointer(c_uint64(0))
    max_value = pointer(c_uint64(0))
    res = _secp.secp256k1_rangeproof_verify(
        context,
        min_value,
        max_value,
        value_commitment,
        proof,
        len(proof),
        script_pubkey,
        len(script_pubkey),
        generator,
    )
    if res != 1:
        raise RuntimeError("Failed to verify the proof")
    return min_value.contents.value, max_value.contents.value


@locked
def rangeproof_sign(
    nonce,
    value,
    value_commitment,
    vbf,
    message,
    extra,
    gen,
    min_value=1,
    exp=0,
    min_bits=52,
    context=_secp.ctx,
):
    if value == 0:
        min_value = 0
    if len(gen) != 64:
        raise ValueError("Generator should be 64 bytes long")
    if len(nonce) != 32:
        raise ValueError("Nonce should be 32 bytes long")
    if len(value_commitment) != 64:
        raise ValueError("Value commitment should be 64 bytes long")
    if len(vbf) != 32:
        raise ValueError("Value blinding factor should be 32 bytes long")
    proof = bytes(5134)
    pointer = POINTER(c_uint64)
    prooflen = pointer(c_uint64(len(proof)))
    res = _secp.secp256k1_rangeproof_sign(
        context,
        proof,
        prooflen,
        min_value,
        value_commitment,
        vbf,
        nonce,
        exp,
        min_bits,
        value,
        message,
        len(message),
        extra,
        len(extra),
        gen,
    )
    if res != 1:
        raise RuntimeError("Failed to generate the proof")
    return bytes(proof[: prooflen.contents.value])


@locked
def musig_pubkey_combine(*args, context=_secp.ctx):
    pub = bytes(64)
    # TODO: strange that behaviour is different from pubkey_combine...
    pubkeys = b"".join(args)  # (c_char_p * len(args))(*args)
    res = _secp.secp256k1_musig_pubkey_combine(
        context, None, pub, None, pubkeys, len(args)
    )
    if res == 0:
        raise ValueError("Failed to combine pubkeys")
    return pub


# surjection proof
@locked
def surjectionproof_initialize(
    in_tags, out_tag, seed, tags_to_use=None, iterations=100, context=_secp.ctx
):
    if tags_to_use is None:
        tags_to_use = min(3, len(in_tags))
    if seed is None:
        seed = os.urandom(32)
    proof = bytes(4 + 8 + 256 // 8 + 32 * 257)
    pointer = POINTER(c_size_t)
    input_index = pointer(c_size_t(0))
    input_tags = b"".join(in_tags)
    res = _secp.secp256k1_surjectionproof_initialize(
        context,
        proof,
        input_index,
        input_tags,
        len(in_tags),
        tags_to_use,
        out_tag,
        iterations,
        seed,
    )
    if res == 0:
        raise RuntimeError("Failed to initialize the proof")
    return proof, input_index.contents.value


@locked
def surjectionproof_generate(
    proof, in_idx, in_tags, out_tag, in_abf, out_abf, context=_secp.ctx
):
    res = _secp.secp256k1_surjectionproof_generate(
        context,
        proof,
        b"".join(in_tags),
        len(in_tags),
        out_tag,
        in_idx,
        in_abf,
        out_abf,
    )
    if not res:
        raise RuntimeError("Failed to generate surjection proof")
    return proof


@locked
def surjectionproof_verify(proof, in_tags, out_tag, context=_secp.ctx):
    res = _secp.secp256k1_surjectionproof_verify(
        context, proof, b"".join(in_tags), len(in_tags), out_tag
    )
    return bool(res)


@locked
def surjectionproof_serialize(proof, context=_secp.ctx):
    s = _secp.secp256k1_surjectionproof_serialized_size(context, proof)
    b = bytes(s)
    pointer = POINTER(c_size_t)
    sz = pointer(c_size_t(s))
    _secp.secp256k1_surjectionproof_serialize(context, b, sz, proof)
    if s != sz.contents.value:
        raise RuntimeError("Failed to serialize surjection proof - size mismatch")
    return b


@locked
def surjectionproof_parse(proof, context=_secp.ctx):
    parsed_proof = bytes(4 + 8 + 256 // 8 + 32 * 257)
    res = _secp.secp256k1_surjectionproof_parse(
        context, parsed_proof, proof, len(proof)
    )
    if res == 0:
        raise RuntimeError("Failed to parse surjection proof")
    return parsed_proof



# Copyright (c) 2021 Pieter Wuille
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.
"""Pure Python RIPEMD160 implementation."""

# Message schedule indexes for the left path.
ML = [
    0,
    1,
    2,
    3,
    4,
    5,
    6,
    7,
    8,
    9,
    10,
    11,
    12,
    13,
    14,
    15,
    7,
    4,
    13,
    1,
    10,
    6,
    15,
    3,
    12,
    0,
    9,
    5,
    2,
    14,
    11,
    8,
    3,
    10,
    14,
    4,
    9,
    15,
    8,
    1,
    2,
    7,
    0,
    6,
    13,
    11,
    5,
    12,
    1,
    9,
    11,
    10,
    0,
    8,
    12,
    4,
    13,
    3,
    7,
    15,
    14,
    5,
    6,
    2,
    4,
    0,
    5,
    9,
    7,
    12,
    2,
    10,
    14,
    1,
    3,
    8,
    11,
    6,
    15,
    13,
]

# Message schedule indexes for the right path.
MR = [
    5,
    14,
    7,
    0,
    9,
    2,
    11,
    4,
    13,
    6,
    15,
    8,
    1,
    10,
    3,
    12,
    6,
    11,
    3,
    7,
    0,
    13,
    5,
    10,
    14,
    15,
    8,
    12,
    4,
    9,
    1,
    2,
    15,
    5,
    1,
    3,
    7,
    14,
    6,
    9,
    11,
    8,
    12,
    2,
    10,
    0,
    4,
    13,
    8,
    6,
    4,
    1,
    3,
    11,
    15,
    0,
    5,
    12,
    2,
    13,
    9,
    7,
    10,
    14,
    12,
    15,
    10,
    4,
    1,
    5,
    8,
    7,
    6,
    2,
    13,
    14,
    0,
    3,
    9,
    11,
]

# Rotation counts for the left path.
RL = [
    11,
    14,
    15,
    12,
    5,
    8,
    7,
    9,
    11,
    13,
    14,
    15,
    6,
    7,
    9,
    8,
    7,
    6,
    8,
    13,
    11,
    9,
    7,
    15,
    7,
    12,
    15,
    9,
    11,
    7,
    13,
    12,
    11,
    13,
    6,
    7,
    14,
    9,
    13,
    15,
    14,
    8,
    13,
    6,
    5,
    12,
    7,
    5,
    11,
    12,
    14,
    15,
    14,
    15,
    9,
    8,
    9,
    14,
    5,
    6,
    8,
    6,
    5,
    12,
    9,
    15,
    5,
    11,
    6,
    8,
    13,
    12,
    5,
    12,
    13,
    14,
    11,
    8,
    5,
    6,
]

# Rotation counts for the right path.
RR = [
    8,
    9,
    9,
    11,
    13,
    15,
    15,
    5,
    7,
    7,
    8,
    11,
    14,
    14,
    12,
    6,
    9,
    13,
    15,
    7,
    12,
    8,
    9,
    11,
    7,
    7,
    12,
    7,
    6,
    15,
    13,
    11,
    9,
    7,
    15,
    11,
    8,
    6,
    6,
    14,
    12,
    13,
    5,
    14,
    13,
    13,
    7,
    5,
    15,
    5,
    8,
    11,
    14,
    14,
    6,
    14,
    6,
    9,
    12,
    9,
    12,
    5,
    15,
    8,
    8,
    5,
    12,
    9,
    12,
    5,
    14,
    6,
    8,
    13,
    6,
    5,
    15,
    13,
    11,
    11,
]

# K constants for the left path.
KL = [0, 0x5A827999, 0x6ED9EBA1, 0x8F1BBCDC, 0xA953FD4E]

# K constants for the right path.
KR = [0x50A28BE6, 0x5C4DD124, 0x6D703EF3, 0x7A6D76E9, 0]


def fi(x, y, z, i):
    """The f1, f2, f3, f4, and f5 functions from the specification."""
    if i == 0:
        return x ^ y ^ z
    elif i == 1:
        return (x & y) | (~x & z)
    elif i == 2:
        return (x | ~y) ^ z
    elif i == 3:
        return (x & z) | (y & ~z)
    elif i == 4:
        return x ^ (y | ~z)
    else:
        assert False


def rol(x, i):
    """Rotate the bottom 32 bits of x left by i bits."""
    return ((x << i) | ((x & 0xFFFFFFFF) >> (32 - i))) & 0xFFFFFFFF


def compress(h0, h1, h2, h3, h4, block):
    """Compress state (h0, h1, h2, h3, h4) with block."""
    # Left path variables.
    al, bl, cl, dl, el = h0, h1, h2, h3, h4
    # Right path variables.
    ar, br, cr, dr, er = h0, h1, h2, h3, h4
    # Message variables.
    x = [int.from_bytes(block[4 * i : 4 * (i + 1)], "little") for i in range(16)]

    # Iterate over the 80 rounds of the compression.
    for j in range(80):
        rnd = j >> 4
        # Perform left side of the transformation.
        al = rol(al + fi(bl, cl, dl, rnd) + x[ML[j]] + KL[rnd], RL[j]) + el
        al, bl, cl, dl, el = el, al, bl, rol(cl, 10), dl
        # Perform right side of the transformation.
        ar = rol(ar + fi(br, cr, dr, 4 - rnd) + x[MR[j]] + KR[rnd], RR[j]) + er
        ar, br, cr, dr, er = er, ar, br, rol(cr, 10), dr

    # Compose old state, left transform, and right transform into new state.
    return h1 + cl + dr, h2 + dl + er, h3 + el + ar, h4 + al + br, h0 + bl + cr


def ripemd160(data):
    """Compute the RIPEMD-160 hash of data."""
    # Initialize state.
    state = (0x67452301, 0xEFCDAB89, 0x98BADCFE, 0x10325476, 0xC3D2E1F0)
    # Process full 64-byte blocks in the input.
    for b in range(len(data) >> 6):
        state = compress(*state, data[64 * b : 64 * (b + 1)])
    # Construct final blocks (with padding and size).
    pad = b"\x80" + b"\x00" * ((119 - len(data)) & 63)
    fin = data[len(data) & ~63 :] + pad + (8 * len(data)).to_bytes(8, "little")
    # Process final blocks.
    for b in range(len(fin) >> 6):
        state = compress(*state, fin[64 * b : 64 * (b + 1)])
    # Produce output.
    return b"".join((h & 0xFFFFFFFF).to_bytes(4, "little") for h in state)



from typing import List, Iterable, Mapping

from .common import write_varint, sha256

NIL = bytes([0] * 32)


def floor_lg(n: int) -> int:
    """Return floor(log_2(n)) for a positive integer `n`"""

    assert n > 0

    r = 0
    t = 1
    while 2 * t <= n:
        t = 2 * t
        r = r + 1
    return r


def ceil_lg(n: int) -> int:
    """Return ceiling(log_2(n)) for a positive integer `n`."""

    assert n > 0

    r = 0
    t = 1
    while t < n:
        t = 2 * t
        r = r + 1
    return r


def is_power_of_2(n: int) -> bool:
    """For a positive integer `n`, returns `True` is `n` is a perfect power of 2, `False` otherwise."""

    assert n >= 1

    return n & (n - 1) == 0


def largest_power_of_2_less_than(n: int) -> int:
    """For an integer `n` which is at least 2, returns the largest exact power of 2 that is strictly less than `n`."""

    assert n > 1

    if is_power_of_2(n):
        return n // 2
    else:
        return 1 << floor_lg(n)


def element_hash(element_preimage: bytes) -> bytes:
    """Computes the hash of an element to be stored in the Merkle tree."""

    return sha256(b'\x00' + element_preimage)


def combine_hashes(left: bytes, right: bytes) -> bytes:
    if len(left) != 32 or len(right) != 32:
        raise ValueError("The elements must be 32-bytes sha256 outputs.")

    return sha256(b'\x01' + left + right)


# root is the only node with parent == None
# leaves have left == right == None
class Node:
    def __init__(self, left, right, parent, value: bytes):
        self.left = left
        self.right = right
        self.parent = parent
        self.value = value

    def recompute_value(self):
        assert self.left is not None
        assert self.right is not None
        self.value = combine_hashes(self.left.value, self.right.value)

    def sibling(self):
        if self.parent is None:
            raise IndexError("The root does not have a sibling.")

        if self.parent.left == self:
            return self.parent.right
        elif self.parent.right == self:
            return self.parent.left
        else:
            raise IndexError("Invalid state: not a child of his parent.")


def make_tree(leaves: List[Node], begin: int, size: int) -> Node:
    """Given a list of nodes, builds the left-complete Merkle tree on top of it.
    The nodes in `leaves` are modified by setting their `parent` field appropriately.
    It returns the root of the newly built tree.
    """

    if size == 0:
        return []
    if size == 1:
        return leaves[begin]

    lchild_size = largest_power_of_2_less_than(size)

    lchild = make_tree(leaves, begin, lchild_size)
    rchild = make_tree(leaves, begin + lchild_size, size - lchild_size)
    root = Node(lchild, rchild, None, None)
    root.recompute_value()
    lchild.parent = rchild.parent = root
    return root


class MerkleTree:
    """
    Maintains a dynamic vector of values and the Merkle tree built on top of it. The elements of the vector are stored
    as the leaves of a binary tree. It is possible to add a new element to the vector, or change an existing element;
    the hashes in the Merkle tree will be recomputed after each operation in O(log n) time, for a vector with n
    elements.
    The value of each internal node is the hash of the concatenation of:
    - a single byte 0x01;
    - the values of the left child;
    - the value of the right child.

    The binary tree has the following properties (assuming the vector contains n leaves):
    - There are always n - 1 internal nodes; all the internal nodes have exactly two children.
    - If a subtree has n > 1 leaves, then the left subchild is a complete subtree with p leaves, where p is the largest
      power of 2 smaller than n.
    """

    def __init__(self, elements: Iterable[bytes] = []):
        self.leaves = [Node(None, None, None, el) for el in elements]
        n_elements = len(self.leaves)
        if n_elements > 0:
            self.root_node = make_tree(self.leaves, 0, n_elements)
            self.depth = ceil_lg(n_elements)
        else:
            self.root_node = None
            self.depth = None

    def __len__(self) -> int:
        """Return the total number of leaves in the tree."""
        return len(self.leaves)

    @property
    def root(self) -> bytes:
        """Return the Merkle root, or None if the tree is empty."""
        return NIL if self.root_node is None else self.root_node.value

    def copy(self):
        """Return an identical copy of this Merkle tree."""
        return MerkleTree([leaf.value for leaf in self.leaves])

    def add(self, x: bytes) -> None:
        """Add an element as new leaf, and recompute the tree accordingly. Cost O(log n)."""

        if len(x) != 32:
            raise ValueError("Inserted elements must be exactly 32 bytes long")

        new_leaf = Node(None, None, None, x)
        self.leaves.append(new_leaf)
        if len(self.leaves) == 1:
            self.root_node = new_leaf
            self.depth = 0
            return

        # add a new leaf
        if self.depth == 0:
            ltree_size = 0
        else:
            # number of leaves of the left subtree of cur_root
            ltree_size = 1 << (self.depth - 1)

        cur_root = self.root_node
        cur_root_size = len(self.leaves) - 1

        while not is_power_of_2(cur_root_size):
            cur_root = cur_root.right
            cur_root_size -= ltree_size
            ltree_size /= 2

        # node value will be computed later
        new_node = Node(cur_root, new_leaf, cur_root.parent, None)
        if cur_root.parent is None:
            # replacing the root
            self.depth += 1
            self.root_node = new_node
        else:
            assert cur_root.parent.right == cur_root
            cur_root.parent.right = new_node
        cur_root.parent = new_node
        new_leaf.parent = new_node

        self.fix_up(new_node)

    def set(self, index: int, x: bytes) -> None:
        """
        Set the value of the leaf at position `index` to `x`, recomputing the tree accordingly.
        If `index` equals the current number of leaves, then it is equivalent to `add(x)`.

        Cost: Worst case O(log n).
        """
        assert 0 <= index <= len(self.leaves)

        if not (0 <= index <= len(self.leaves)):
            raise ValueError(
                "The index must be at least 0, and at most the current number of leaves.")

        if len(x) != 32:
            raise ValueError("Inserted elements must be exactly 32 bytes long.")

        if index == len(self.leaves):
            self.add(x)
        else:
            self.leaves[index].value = x
            self.fix_up(self.leaves[index].parent)

    def fix_up(self, node: Node):
        while node is not None:
            node.recompute_value()
            node = node.parent

    def get(self, i: int) -> bytes:
        """Return the value of the leaf with index `i`, where 0 <= i < len(self)."""
        return self.leaves[i].value

    def leaf_index(self, x: bytes) -> int:
        """Return the index of the leaf with hash `x`. Raises `ValueError` if not found."""
        idx = 0
        while idx < len(self):
            if self.leaves[idx].value == x:
                return idx
            idx += 1
        raise ValueError("Leaf not found")

    def prove_leaf(self, index: int) -> List[bytes]:
        """Produce the Merkle proof of membership for the leaf with the given index where 0 <= index < len(self)."""
        node = self.leaves[index]
        proof = []
        while node.parent is not None:
            sibling = node.sibling()
            assert sibling is not None

            proof.append(sibling.value)

            node = node.parent

        return proof


def get_merkleized_map_commitment(mapping: Mapping[bytes, bytes]) -> bytes:
    """Returns a serialized Merkleized map commitment, encoded as the concatenation of:
       - the number of key/value pairs, as a Bitcoin-style varint;
       - the root of the Merkle tree of the keys
       - the root of the Merkle tree of the values.
    """

    items_sorted = list(sorted(mapping.items()))
    keys_hashes = [element_hash(i[0]) for i in items_sorted]
    values_hashes = [element_hash(i[1]) for i in items_sorted]
    return write_varint(len(mapping)) + MerkleTree(keys_hashes).root + MerkleTree(values_hashes).root



import enum
from typing import List, Tuple, Mapping, Union, Iterator, Optional

from .common import bip32_path_from_string, write_varint
from .merkle import get_merkleized_map_commitment, MerkleTree, element_hash
from .wallet import WalletPolicy

# p2 encodes the protocol version implemented
CURRENT_PROTOCOL_VERSION = 1

def chunkify(data: bytes, chunk_len: int) -> Iterator[Tuple[bool, bytes]]:
    size: int = len(data)

    if size <= chunk_len:
        yield True, data
        return

    chunk: int = size // chunk_len
    remaining: int = size % chunk_len
    offset: int = 0

    for i in range(chunk):
        yield False, data[offset: offset + chunk_len]
        offset += chunk_len

    if remaining:
        yield True, data[offset:]


class DefaultInsType(enum.IntEnum):
    GET_VERSION = 0x01

class BitcoinInsType(enum.IntEnum):
    GET_EXTENDED_PUBKEY = 0x00
    REGISTER_WALLET = 0x02
    GET_WALLET_ADDRESS = 0x03
    SIGN_PSBT = 0x04
    GET_MASTER_FINGERPRINT = 0x05
    SIGN_MESSAGE = 0x10

class FrameworkInsType(enum.IntEnum):
    CONTINUE_INTERRUPTED = 0x01


class BitcoinCommandBuilder:
    """APDU command builder for the Bitcoin application."""

    CLA_DEFAULT: int = 0xB0
    CLA_BITCOIN: int = 0xE1
    CLA_FRAMEWORK: int = 0xF8

    def serialize(
        self,
        cla: int,
        ins: Union[int, enum.IntEnum],
        p1: int = 0,
        p2: int = CURRENT_PROTOCOL_VERSION,
        cdata: bytes = b"",
    ) -> dict:
        """Serialize the whole APDU command (header + data).

        Parameters
        ----------
        cla : int
            Instruction class: CLA (1 byte)
        ins : Union[int, IntEnum]
            Instruction code: INS (1 byte)
        p1 : int
            Instruction parameter 1: P1 (1 byte).
        p2 : int
            Instruction parameter 2: P2 (1 byte).
        cdata : bytes
            Bytes of command data.

        Returns
        -------
        dict
            Dictionary representing the APDU message.

        """

        return {"cla": cla, "ins": ins, "p1": p1, "p2": p2, "data": cdata}

    def get_extended_pubkey(self, bip32_path: str, display: bool = False):
        bip32_path: List[bytes] = bip32_path_from_string(bip32_path)

        cdata: bytes = b"".join([
            b'\1' if display else b'\0',
            len(bip32_path).to_bytes(1, byteorder="big"),
            *bip32_path
        ])

        return self.serialize(
            cla=self.CLA_BITCOIN,
            ins=BitcoinInsType.GET_EXTENDED_PUBKEY,
            cdata=cdata,
        )

    def register_wallet(self, wallet: WalletPolicy):
        wallet_bytes = wallet.serialize()

        return self.serialize(
            cla=self.CLA_BITCOIN,
            ins=BitcoinInsType.REGISTER_WALLET,
            cdata=write_varint(len(wallet_bytes)) + wallet_bytes,
        )

    def get_wallet_address(
        self,
        wallet: WalletPolicy,
        wallet_hmac: Optional[bytes],
        address_index: int,
        change: bool,
        display: bool,
    ):
        cdata: bytes = b"".join(
            [
                b'\1' if display else b'\0',                            # 1 byte
                wallet.id,                                              # 32 bytes
                wallet_hmac if wallet_hmac is not None else b'\0' * 32, # 32 bytes
                b"\1" if change else b"\0",                             # 1 byte
                address_index.to_bytes(4, byteorder="big"),             # 4 bytes
            ]
        )

        return self.serialize(
            cla=self.CLA_BITCOIN,
            ins=BitcoinInsType.GET_WALLET_ADDRESS,
            cdata=cdata,
        )

    def sign_psbt(
        self,
        global_mapping: Mapping[bytes, bytes],
        input_mappings: List[Mapping[bytes, bytes]],
        output_mappings: List[Mapping[bytes, bytes]],
        wallet: WalletPolicy,
        wallet_hmac: Optional[bytes],
    ):

        cdata = bytearray()
        cdata += get_merkleized_map_commitment(global_mapping)

        cdata += write_varint(len(input_mappings))
        cdata += MerkleTree(
            [
                element_hash(get_merkleized_map_commitment(m_in))
                for m_in in input_mappings
            ]
        ).root

        cdata += write_varint(len(output_mappings))
        cdata += MerkleTree(
            [
                element_hash(get_merkleized_map_commitment(m_out))
                for m_out in output_mappings
            ]
        ).root

        cdata += wallet.id
        cdata += wallet_hmac if wallet_hmac is not None else b'\0' * 32

        return self.serialize(
            cla=self.CLA_BITCOIN, ins=BitcoinInsType.SIGN_PSBT, cdata=bytes(cdata)
        )

    def get_master_fingerprint(self):
        return self.serialize(
            cla=self.CLA_BITCOIN,
            ins=BitcoinInsType.GET_MASTER_FINGERPRINT
        )

    def sign_message(self, message: bytes, bip32_path: str):
        cdata = bytearray()

        bip32_path: List[bytes] = bip32_path_from_string(bip32_path)

        # split message in 64-byte chunks (last chunk can be smaller)
        n_chunks = (len(message) + 63) // 64
        chunks = [message[64 * i: 64 * i + 64] for i in range(n_chunks)]

        cdata += len(bip32_path).to_bytes(1, byteorder="big")
        cdata += b''.join(bip32_path)

        cdata += write_varint(len(message))

        cdata += MerkleTree(element_hash(c) for c in chunks).root

        return self.serialize(
            cla=self.CLA_BITCOIN,
            ins=BitcoinInsType.SIGN_MESSAGE,
            cdata=bytes(cdata)
        )

    def continue_interrupted(self, cdata: bytes):
        """Command builder for CONTINUE.

        Returns
        -------
        bytes
            APDU command for CONTINUE.

        """
        return self.serialize(
            cla=self.CLA_FRAMEWORK,
            ins=FrameworkInsType.CONTINUE_INTERRUPTED,
            cdata=cdata,
        )



# Original version: https://github.com/bitcoin-core/HWI/blob/3fe369d0379212fae1c72729a179d133b0adc872/hwilib/key.py

#!/usr/bin/env python3
# Copyright (c) 2020 The HWI developers
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.

"""
Key Classes and Utilities
*************************

Classes and utilities for working with extended public keys, key origins, and other key related things.
"""

from . import _base58 as base58
from .common import (
    AddressType,
    Chain,
    hash256,
    hash160,
)
from .errors import BadArgumentError

import binascii
import hmac
import hashlib
import struct
from typing import (
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
)


HARDENED_FLAG = 1 << 31

p = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFC2F
n = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141
G = (0x79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798, 0x483ADA7726A3C4655DA4FBFC0E1108A8FD17B448A68554199C47D08FFB10D4B8)

Point = Optional[Tuple[int, int]]

def H_(x: int) -> int:
    """
    Shortcut function that "hardens" a number in a BIP44 path.
    """
    return x | HARDENED_FLAG

def is_hardened(i: int) -> bool:
    """
    Returns whether an index is hardened
    """
    return i & HARDENED_FLAG != 0


def point_add(p1: Point, p2: Point) -> Point:
    if (p1 is None):
        return p2
    if (p2 is None):
        return p1
    if (p1[0] == p2[0] and p1[1] != p2[1]):
        return None
    if (p1 == p2):
        lam = (3 * p1[0] * p1[0] * pow(2 * p1[1], p - 2, p)) % p
    else:
        lam = ((p2[1] - p1[1]) * pow(p2[0] - p1[0], p - 2, p)) % p
    x3 = (lam * lam - p1[0] - p2[0]) % p
    return (x3, (lam * (p1[0] - x3) - p1[1]) % p)


def point_mul(p: Point, n: int) -> Point:
    r = None
    for i in range(256):
        if ((n >> i) & 1):
            r = point_add(r, p)
        p = point_add(p, p)
    return r


def deserialize_point(b: bytes) -> Point:
    x = int.from_bytes(b[1:], byteorder="big")
    y = pow((x * x * x + 7) % p, (p + 1) // 4, p)
    if (y & 1 != b[0] & 1):
        y = p - y
    return (x, y)


def bytes_to_point(point_bytes: bytes) -> Point:
    header = point_bytes[0]
    if header == 4:
        x = point_bytes = point_bytes[1:33]
        y = point_bytes = point_bytes[33:65]
        return (int(binascii.hexlify(x), 16), int(binascii.hexlify(y), 16))
    return deserialize_point(point_bytes)

def point_to_bytes(p: Point) -> bytes:
    if p is None:
        raise ValueError("Cannot convert None to bytes")
    return (b'\x03' if p[1] & 1 else b'\x02') + p[0].to_bytes(32, byteorder="big")

def int_from_bytes(b: bytes) -> int:
    return int(binascii.hexlify(b), 16)

def lift_x(x: int) -> 'Point':
    c = (pow(x, 3, p) + 7) % p
    y = pow(c, (p + 1) // 4, p)

    assert(c == y * y % p)

    return (x, p - y if y & 1 else y)

def tagged_hash(tag: str, data: bytes) -> bytes:
    hashtag = hashlib.sha256(tag.encode()).digest()
    return hashlib.sha256(hashtag + hashtag + data).digest()


def taproot_tweak_pubkey(pubkey: bytes, h: bytes) -> Tuple[int, bytes]:
    t = int_from_bytes(tagged_hash("TapTweak", pubkey + h))
    if t >= p:
        raise ValueError
    Q = point_add(lift_x(int_from_bytes(pubkey)), point_mul(G, t))
    return 0 if Q[1] & 1 == 0 else 1, Q[0].to_bytes(32, byteorder="big")


def get_taproot_output_key(derived_key: bytes) -> bytes:
    assert(len(derived_key) == 33)
    p = derived_key[1:]
    _, key = taproot_tweak_pubkey(p, b'')
    return key


# An extended public key (xpub) or private key (xprv). Just a data container for now.
# Only handles deserialization of extended keys into component data to be handled by something else
class ExtendedKey(object):
    """
    A BIP 32 extended public key.
    """

    MAINNET_PUBLIC = b'\x04\x88\xB2\x1E'
    MAINNET_PRIVATE = b'\x04\x88\xAD\xE4'
    TESTNET_PUBLIC = b'\x04\x35\x87\xCF'
    TESTNET_PRIVATE = b'\x04\x35\x83\x94'

    def __init__(self, version: bytes, depth: int, parent_fingerprint: bytes, child_num: int, chaincode: bytes, privkey: Optional[bytes], pubkey: bytes) -> None:
        """
        :param version: The version bytes for this xpub
        :param depth: The depth of this xpub as defined in BIP 32
        :param parent_fingerprint: The 4 byte fingerprint of the parent xpub as defined in BIP 32
        :param child_num: The number of this xpub as defined in BIP 32
        :param chaincode: The chaincode of this xpub as defined in BIP 32
        :param privkey: The private key for this xpub if available
        :param pubkey: The public key for this xpub
        """
        self.version: bytes = version
        self.is_testnet: bool = version == ExtendedKey.TESTNET_PUBLIC or version == ExtendedKey.TESTNET_PRIVATE
        self.is_private: bool = version == ExtendedKey.MAINNET_PRIVATE or version == ExtendedKey.TESTNET_PRIVATE
        self.depth: int = depth
        self.parent_fingerprint: bytes = parent_fingerprint
        self.child_num: int = child_num
        self.chaincode: bytes = chaincode
        self.pubkey: bytes = pubkey
        self.privkey: Optional[bytes] = privkey

    @classmethod
    def deserialize(cls, xpub: str) -> 'ExtendedKey':
        """
        Create an :class:`~ExtendedKey` from a Base58 check encoded xpub

        :param xpub: The Base58 check encoded xpub
        """
        data = base58.decode(xpub)[:-4] # Decoded xpub without checksum
        return cls.from_bytes(data)

    @classmethod
    def from_bytes(cls, data: bytes) -> 'ExtendedKey':
        """
        Create an :class:`~ExtendedKey` from a serialized xpub

        :param xpub: The serialized xpub
        """

        version = data[0:4]
        if version not in [ExtendedKey.MAINNET_PRIVATE, ExtendedKey.MAINNET_PUBLIC, ExtendedKey.TESTNET_PRIVATE, ExtendedKey.TESTNET_PUBLIC]:
            raise BadArgumentError(f"Extended key magic of {version.hex()} is invalid")
        is_private = version == ExtendedKey.MAINNET_PRIVATE or version == ExtendedKey.TESTNET_PRIVATE
        depth = data[4]
        parent_fingerprint = data[5:9]
        child_num = struct.unpack('>I', data[9:13])[0]
        chaincode = data[13:45]

        if is_private:
            privkey = data[46:]
            pubkey = point_to_bytes(point_mul(G, int.from_bytes(privkey, byteorder="big")))
            return cls(version, depth, parent_fingerprint, child_num, chaincode, privkey, pubkey)
        else:
            pubkey = data[45:78]
            return cls(version, depth, parent_fingerprint, child_num, chaincode, None, pubkey)

    def serialize(self) -> bytes:
        """
        Serialize the ExtendedKey with the serialization format described in BIP 32.
        Does not create an xpub string, but the bytes serialized here can be Base58 check encoded into one.

        :return: BIP 32 serialized extended key
        """
        r = self.version + struct.pack('B', self.depth) + self.parent_fingerprint + struct.pack('>I', self.child_num) + self.chaincode
        if self.is_private:
            if self.privkey is None:
                raise ValueError("Somehow we are private but don't have a privkey")
            r += b"\x00" + self.privkey
        else:
            r += self.pubkey
        return r

    def to_string(self) -> str:
        """
        Serialize the ExtendedKey as a Base58 check encoded xpub string

        :return: Base58 check encoded xpub
        """
        data = self.serialize()
        checksum = hash256(data)[0:4]
        return base58.encode(data + checksum)

    def get_printable_dict(self) -> Dict[str, object]:
        """
        Get the attributes of this ExtendedKey as a dictionary that can be printed

        :return: Dictionary containing ExtendedKey information that can be printed
        """
        d: Dict[str, object] = {}
        d['testnet'] = self.is_testnet
        d['private'] = self.is_private
        d['depth'] = self.depth
        d['parent_fingerprint'] = binascii.hexlify(self.parent_fingerprint).decode()
        d['child_num'] = self.child_num
        d['chaincode'] = binascii.hexlify(self.chaincode).decode()
        if self.is_private and isinstance(self.privkey, bytes):
            d['privkey'] = binascii.hexlify(self.privkey).decode()
        d['pubkey'] = binascii.hexlify(self.pubkey).decode()
        return d

    def derive_priv(self, i: int) -> 'ExtendedKey':
        """
        Derive the private key at the given child index.

        :param i: The child index of the pubkey to derive
        """
        if not self.privkey:
            raise ValueError("Can only derive a private key from an extended private key")

        # Data to HMAC
        if is_hardened(i):
            data = b'\0' + self.privkey + struct.pack(">L", i)
        else:
            data = self.pubkey + struct.pack(">L", i)

        # Get HMAC of data
        Ihmac = hmac.new(self.chaincode, data, hashlib.sha512).digest()
        Il = Ihmac[:32]
        Ir = Ihmac[32:]

        # Construct new key material from Il and current private key
        Il_int = int.from_bytes(Il, byteorder="big")
        if Il_int > n:
            return None

        privkey_int = int.from_bytes(self.privkey, byteorder="big")
        k_int = (Il_int + privkey_int) % n
        if (k_int == 0):
            return None

        privkey = k_int.to_bytes(32, byteorder="big")
        pubkey = point_to_bytes(point_mul(G, k_int))

        chaincode = Ir
        fingerprint = hash160(self.pubkey)[0:4]
        return ExtendedKey(ExtendedKey.TESTNET_PRIVATE if self.is_testnet else ExtendedKey.MAINNET_PRIVATE, self.depth + 1, fingerprint, i, chaincode, privkey, pubkey)

    def derive_pub(self, i: int) -> 'ExtendedKey':
        """
        Derive the public key at the given child index.

        :param i: The child index of the pubkey to derive
        """
        if is_hardened(i):
            raise ValueError("Index cannot be larger than 2^31")

        # Data to HMAC.  Same as CKDpriv() for public child key.
        data = self.pubkey + struct.pack(">L", i)

        # Get HMAC of data
        Ihmac = hmac.new(self.chaincode, data, hashlib.sha512).digest()
        Il = Ihmac[:32]
        Ir = Ihmac[32:]

        # Construct curve point Il*G+K
        Il_int = int(binascii.hexlify(Il), 16)
        child_pubkey = point_add(point_mul(G, Il_int), bytes_to_point(self.pubkey))

        # Construct and return a new BIP32Key
        pubkey = point_to_bytes(child_pubkey)
        chaincode = Ir
        fingerprint = hash160(self.pubkey)[0:4]
        return ExtendedKey(ExtendedKey.TESTNET_PUBLIC if self.is_testnet else ExtendedKey.MAINNET_PUBLIC, self.depth + 1, fingerprint, i, chaincode, None, pubkey)

    def derive_priv_path(self, path: Sequence[int]) -> 'ExtendedKey':
        """
        Derive the private key at the given path

        :param path: Sequence of integers for the path of the key to derive
        """
        key = self
        for i in path:
            key = key.derive_priv(i)
        return key

    def derive_pub_path(self, path: Sequence[int]) -> 'ExtendedKey':
        """
        Derive the public key at the given path

        :param path: Sequence of integers for the path of the pubkey to derive
        """
        key = self
        for i in path:
            key = key.derive_pub(i)
        return key

    def neutered(self) -> 'ExtendedKey':
        """
        Returns the public key corresponding to this private key.
        """
        if not self.is_private:
            raise ValueError("It is already a public key")

        return ExtendedKey(ExtendedKey.TESTNET_PUBLIC if self.is_testnet else ExtendedKey.MAINNET_PUBLIC, self.depth, self.parent_fingerprint, self.child_num, self.chaincode, None, self.pubkey)


class KeyOriginInfo(object):
    """
    Object representing the origin of a key.
    """

    def __init__(self, fingerprint: bytes, path: Sequence[int]) -> None:
        """
        :param fingerprint: The 4 byte BIP 32 fingerprint of a parent key from which this key is derived from
        :param path: The derivation path to reach this key from the key at ``fingerprint``
        """
        self.fingerprint: bytes = fingerprint
        self.path: Sequence[int] = path

    @classmethod
    def deserialize(cls, s: bytes) -> 'KeyOriginInfo':
        """
        Deserialize a serialized KeyOriginInfo.
        They will be serialized in the same way that PSBTs serialize derivation paths
        """
        fingerprint = s[0:4]
        s = s[4:]
        path = list(struct.unpack("<" + "I" * (len(s) // 4), s))
        return cls(fingerprint, path)

    def serialize(self) -> bytes:
        """
        Serializes the KeyOriginInfo in the same way that derivation paths are stored in PSBTs
        """
        r = self.fingerprint
        r += struct.pack("<" + "I" * len(self.path), *self.path)
        return r

    def _path_string(self) -> str:
        s = ""
        for i in self.path:
            hardened = is_hardened(i)
            i &= ~HARDENED_FLAG
            s += "/" + str(i)
            if hardened:
                s += "h"
        return s

    def to_string(self) -> str:
        """
        Return the KeyOriginInfo as a string in the form <fingerprint>/<index>/<index>/...
        This is the same way that KeyOriginInfo is shown in descriptors
        """
        s = binascii.hexlify(self.fingerprint).decode()
        s += self._path_string()
        return s

    @classmethod
    def from_string(cls, s: str) -> 'KeyOriginInfo':
        """
        Create a KeyOriginInfo from the string

        :param s: The string to parse
        """
        s = s.lower()
        entries = s.split("/")
        fingerprint = binascii.unhexlify(s[0:8])
        path: Sequence[int] = []
        if len(entries) > 1:
            path = parse_path(s[9:])
        return cls(fingerprint, path)

    def get_derivation_path(self) -> str:
        """
        Return the string for just the path
        """
        return "m" + self._path_string()

    def get_full_int_list(self) -> List[int]:
        """
        Return a list of ints representing this KeyOriginInfo.
        The first int is the fingerprint, followed by the path
        """
        xfp = [struct.unpack("<I", self.fingerprint)[0]]
        xfp.extend(self.path)
        return xfp


def parse_path(nstr: str) -> List[int]:
    """
    Convert BIP32 path string to list of uint32 integers with hardened flags.
    Several conventions are supported to set the hardened flag: -1, 1', 1h

    e.g.: "0/1h/1" -> [0, 0x80000001, 1]

    :param nstr: path string
    :return: list of integers
    """
    if not nstr:
        return []

    n = nstr.split("/")

    # m/a/b/c => a/b/c
    if n[0] == "m":
        n = n[1:]

    def str_to_harden(x: str) -> int:
        if x.startswith("-"):
            return H_(abs(int(x)))
        elif x.endswith(("h", "'")):
            return H_(int(x[:-1]))
        else:
            return int(x)

    try:
        return [str_to_harden(x) for x in n]
    except Exception:
        raise ValueError("Invalid BIP32 path", nstr)


def get_bip44_purpose(addrtype: AddressType) -> int:
    """
    Determine the BIP 44 purpose based on the given :class:`~hwilib.common.AddressType`.

    :param addrtype: The address type
    """
    if addrtype == AddressType.LEGACY:
        return 44
    elif addrtype == AddressType.SH_WIT:
        return 49
    elif addrtype == AddressType.WIT:
        return 84
    elif addrtype == AddressType.TAP:
        return 86
    else:
        raise ValueError("Unknown address type")


def get_bip44_chain(chain: Chain) -> int:
    """
    Determine the BIP 44 coin type based on the Bitcoin chain type.

    For the Bitcoin mainnet chain, this returns 0. For the other chains, this returns 1.

    :param chain: The chain
    """
    if chain == Chain.MAIN:
        return 0
    else:
        return 1



"""
*******************************************************************************
*   BTChip Bitcoin Hardware Wallet Python API
*   (c) 2014 BTChip - 1BTChip7VfTnrPra5jqci7ejnMguuHogTn
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*   limitations under the License.
********************************************************************************
"""

from abc import ABCMeta, abstractmethod
from .btchipException import *
from .ledgerWrapper import wrapCommandAPDU, unwrapResponseAPDU
from binascii import hexlify
import time
import os
import struct
import socket

try:
	import hid
	HID = True
except ImportError:
	HID = False

try:
	from smartcard.Exceptions import NoCardException
	from smartcard.System import readers
	from smartcard.util import toHexString, toBytes
	SCARD = True
except ImportError:
	SCARD = False

class DongleWait(object):
	__metaclass__ = ABCMeta

	@abstractmethod
	def waitFirstResponse(self, timeout):
		pass

class Dongle(object):
	__metaclass__ = ABCMeta

	@abstractmethod
	def exchange(self, apdu, timeout=20000):
		pass

	@abstractmethod
	def close(self):
		pass

	def setWaitImpl(self, waitImpl):
		self.waitImpl = waitImpl

class HIDDongleHIDAPI(Dongle, DongleWait):

	def __init__(self, device, ledger=False, debug=False):
		self.device = device
		self.ledger = ledger		
		self.debug = debug
		self.waitImpl = self
		self.opened = True

	def exchange(self, apdu, timeout=20000):
		if self.debug:
			print("=> %s" % hexlify(apdu))
		if self.ledger:
			apdu = wrapCommandAPDU(0x0101, apdu, 64)		
		padSize = len(apdu) % 64
		tmp = apdu
		if padSize != 0:
			tmp.extend([0] * (64 - padSize))
		offset = 0
		while(offset != len(tmp)):
			data = tmp[offset:offset + 64]
			data = bytearray([0]) + data
			self.device.write(data)
			offset += 64
		dataLength = 0
		dataStart = 2		
		result = self.waitImpl.waitFirstResponse(timeout)
		if not self.ledger:
			if result[0] == 0x61: # 61xx : data available
				self.device.set_nonblocking(False)
				dataLength = result[1]
				dataLength += 2
				if dataLength > 62:
					remaining = dataLength - 62
					while(remaining != 0):
						if remaining > 64:
							blockLength = 64
						else:
							blockLength = remaining
						result.extend(bytearray(self.device.read(65))[0:blockLength])
						remaining -= blockLength
				swOffset = dataLength
				dataLength -= 2
				self.device.set_nonblocking(True)
			else:
				swOffset = 0
		else:
			self.device.set_nonblocking(False)
			while True:
				response = unwrapResponseAPDU(0x0101, result, 64)
				if response is not None:
					result = response
					dataStart = 0
					swOffset = len(response) - 2
					dataLength = len(response) - 2
					self.device.set_nonblocking(True)
					break
				result.extend(bytearray(self.device.read(65)))
		sw = (result[swOffset] << 8) + result[swOffset + 1]
		response = result[dataStart : dataLength + dataStart]
		if self.debug:
			print("<= %s%.2x" % (hexlify(response), sw))
		if sw != 0x9000:
			raise BTChipException("Invalid status %04x" % sw, sw)
		return response

	def waitFirstResponse(self, timeout):
		start = time.time()
		data = ""
		while len(data) == 0:
			data = self.device.read(65)
			if not len(data):
				if time.time() - start > timeout:
					raise BTChipException("Timeout")
				time.sleep(0.02)
		return bytearray(data)

	def close(self):
		if self.opened:
			try:
				self.device.close()
			except Exception:
				pass
		self.opened = False

class DongleServer(Dongle):

	def __init__(self, server, port, debug=False):
		self.server = server
		self.port = port
		self.debug = debug
		self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		try:
			self.socket.connect((self.server, self.port))
		except Exception:
			raise BTChipException("Proxy connection failed")

	def exchange(self, apdu, timeout=20000):
		if self.debug:
			print("=> %s" % hexlify(apdu))		
		self.socket.send(struct.pack(">I", len(apdu)))
		self.socket.send(apdu)
		size = struct.unpack(">I", self.socket.recv(4))[0]
		response = self.socket.recv(size)
		sw = struct.unpack(">H", self.socket.recv(2))[0]
		if self.debug:
			print("<= %s%.2x" % (hexlify(response), sw))
		if sw != 0x9000:
			raise BTChipException("Invalid status %04x" % sw, sw)
		return bytearray(response)

	def close(self):
		try:
			self.socket.close()
		except Exception:
			pass



""" Compact Int parsing / serialization """
import io


def to_bytes(i: int) -> bytes:
    """encodes an integer as a compact int"""
    if i < 0:
        raise ValueError("integer can't be negative: {}".format(i))
    order = 0
    while i >> (8 * (2**order)):
        order += 1
    if order == 0:
        if i < 0xFD:
            return bytes([i])
        order = 1
    if order > 3:
        raise ValueError("integer too large: {}".format(i))
    return bytes([0xFC + order]) + i.to_bytes(2**order, "little")


def from_bytes(b: bytes) -> int:
    s = io.BytesIO(b)
    res = read_from(s)
    if len(s.read(1)) > 0:
        raise ValueError("Too many bytes")
    return res


def read_from(stream) -> int:
    """reads a compact integer from a stream"""
    c = stream.read(1)
    if not isinstance(c, bytes):
        raise TypeError("Bytes must be returned from stream.read()")
    if len(c) != 1:
        raise RuntimeError("Can't read one byte from the stream")
    i = c[0]
    if i >= 0xFD:
        bytes_to_read = 2 ** (i - 0xFC)
        return int.from_bytes(stream.read(bytes_to_read), "little")
    else:
        return i
